import numpy as np

from autooc.algorithm.parameters import params
from autooc.stats.stats import stats
from autooc.utilities.stats.trackers import cache, runtime_error_cache


def evaluate_fitness(individuals):
    """
    Evaluate an entire population of individuals. Invalid individuals are given
    a default bad fitness. If params['CACHE'] is specified then individuals
    have their fitness stored in a dictionary called utilities.trackers.cache.
    Dictionary keys are the string of the phenotype.
    There are currently three options for use with the cache:
        1. If params['LOOKUP_FITNESS'] is specified (default case if
           params['CACHE'] is specified), individuals which have already been
           evaluated have their previous fitness read directly from the cache,
           thus saving fitness evaluations.
        2. If params['LOOKUP_BAD_FITNESS'] is specified, individuals which
           have already been evaluated are given a default bad fitness.
        3. If params['MUTATE_DUPLICATES'] is specified, individuals which
           have already been evaluated are mutated to produce new unique
           individuals which have not been encountered yet by the search
           process.

    :param individuals: A population of individuals to be evaluated.
    :return: A population of fully evaluated individuals.
    """

    results, pool = [], None

    if params["MULTICORE"]:
        pool = params["POOL"]

    for name, ind in enumerate(individuals):
        ind.name = name

        # Iterate over all individuals in the population.
        if ind.invalid:
            # Invalid individuals cannot be evaluated and are given a bad
            # default fitness.
            ind.fitness = params["FITNESS_FUNCTION"].default_fitness
            stats["invalids"] += 1

        else:
            eval_ind = True

            # Valid individuals can be evaluated.
            if params["CACHE"] and ind.phenotype in cache:
                # The individual has been encountered before in
                # the utilities.trackers.cache.

                if params["LOOKUP_FITNESS"]:
                    # Set the fitness as the previous fitness from the
                    # cache.
                    ind.fitness = cache[ind.phenotype]
                    eval_ind = False

                elif params["LOOKUP_BAD_FITNESS"]:
                    # Give the individual a bad default fitness.
                    ind.fitness = params["FITNESS_FUNCTION"].default_fitness
                    eval_ind = False

                elif params["MUTATE_DUPLICATES"]:
                    # Mutate the individual to produce a new phenotype
                    # which has not been encountered yet.
                    while (not ind.phenotype) or ind.phenotype in cache:
                        ind = params["MUTATION"](ind)
                        stats["regens"] += 1

                    # Need to overwrite the current individual in the pop.
                    individuals[name] = ind
                    ind.name = name

            if eval_ind:
                results = eval_or_append(ind, results, pool)

    if params["MULTICORE"]:
        for result in results:
            # Execute all jobs in the pool.
            ind = result.get()

            # Set the fitness of the evaluated individual by placing the
            # evaluated individual back into the population.
            individuals[ind.name] = ind

            # Add the evaluated individual to the cache.
            cache[ind.phenotype] = ind.fitness

            # Check if individual had a runtime error.
            if ind.runtime_error:
                runtime_error_cache.append(ind.phenotype)

    return individuals


def eval_or_append(ind, results, pool):
    """
    Evaluates an individual if sequential evaluation is being used. If
    multi-core parallel evaluation is being used, adds the individual to the
    pool to be evaluated.

    :param ind: An individual to be evaluated.
    :param results: A list of individuals to be evaluated by the multicore
    pool of workers.
    :param pool: A pool of workers for multicore evaluation.
    :return: The evaluated individual or the list of individuals to be
    evaluated.
    """

    if params["MULTICORE"]:
        # Add the individual to the pool of jobs.
        results.append(pool.apply_async(ind.evaluate, ()))
        return results

    else:
        # Evaluate the individual.
        ind.evaluate()

        # Check if individual had a runtime error.
        if ind.runtime_error:
            runtime_error_cache.append(ind.phenotype)

        if params["CACHE"]:
            # The phenotype string of the individual does not appear
            # in the cache, it must be evaluated and added to the
            # cache.

            if (
                isinstance(ind.fitness, list)
                and not any([np.isnan(i) for i in ind.fitness])
            ) or (not isinstance(ind.fitness, list) and not np.isnan(ind.fitness)):

                # All fitnesses are valid.
                cache[ind.phenotype] = ind.fitness



from autooc.fitness.base_ff_classes.base_ff import base_ff


class ff_template(base_ff):
    """
    Basic fitness function template for writing new fitness functions. This
    basic template inherits from the base fitness function class, which
    contains various checks and balances.

    Note that all fitness functions must be implemented as a class.

    Note that the class name must be the same as the file name.

    Important points to note about base fitness function class from which
    this template inherits:

      - Default Fitness values (can be referenced as "self.default_fitness")
        are set to NaN in the base class. While this can be over-written,
        PonyGE2 works best when it can filter solutions by NaN values.

      - The standard fitness objective of the base fitness function class is
        to minimise fitness. If the objective is to maximise fitness,
        this can be over-written by setting the flag "maximise = True".

    """

    # The base fitness function class is set up to minimise fitness.
    # However, if you wish to maximise fitness values, you only need to
    # change the "maximise" attribute here to True rather than False.
    # Note that if fitness is being minimised, it is not necessary to
    # re-define/overwrite the maximise attribute here, as it already exists
    # in the base fitness function class.
    maximise = False

    def __init__(self):
        """
        All fitness functions which inherit from the bass fitness function
        class must initialise the base class during their own initialisation.
        """

        # Initialise base fitness function class.
        super().__init__()

    def evaluate(self, ind, **kwargs):
        """
        Default fitness execution call for all fitness functions. When
        implementing a new fitness function, this is where code should be added
        to evaluate target phenotypes.

        There is no need to implement a __call__() method for new fitness
        functions which inherit from the base class; the "evaluate()" function
        provided here allows for this. Implementing a __call__() method for new
        fitness functions will over-write the __call__() method in the base
        class, removing much of the functionality and use of the base class.

        :param ind: An individual to be evaluated.
        :param kwargs: Optional extra arguments.
        :return: The fitness of the evaluated individual.
        """

        # Evaluate the fitness of the phenotype
        fitness = eval(ind.phenotype)

        return fitness



from math import floor
from re import DOTALL, MULTILINE, finditer, match
from sys import maxsize

from autooc.algorithm.parameters import params


class Grammar(object):
    """
    Parser for Backus-Naur Form (BNF) Context-Free Grammars.
    """

    def __init__(self, file_name):
        """
        Initialises an instance of the grammar class. This instance is used
        to parse a given file_name grammar.

        :param file_name: A specified BNF grammar file.
        """

        if file_name.endswith("pybnf"):
            # Use python filter for parsing grammar output as grammar output
            # contains indented python code.
            self.python_mode = True

        else:
            # No need to filter/interpret grammar output, individual
            # phenotypes can be evaluated as normal.
            self.python_mode = False

        # Initialise empty dict for all production rules in the grammar.
        # Initialise empty dict of permutations of solutions possible at
        # each derivation tree depth.
        self.rules, self.permutations = {}, {}

        # Initialise dicts for terminals and non terminals, set params.
        self.non_terminals, self.terminals = {}, {}
        self.start_rule, self.codon_size = None, params["CODON_SIZE"]
        self.min_path, self.max_arity, self.min_ramp = None, None, None

        # Set regular expressions for parsing BNF grammar.
        self.ruleregex = "(?P<rulename><\S+>)\s*::=\s*(?P<production>(?:(?=\#)\#[^\r\n]*|(?!<\S+>\s*::=).+?)+)"
        self.productionregex = (
            "(?=\#)(?:\#.*$)|(?!\#)\s*(?P<production>(?:[^'\"\|\#]+|'.*?'|\".*?\")+)"
        )
        self.productionpartsregex = "\ *([\r\n]+)\ *|([^'\"<\r\n]+)|'(.*?)'|\"(.*?)\"|(?P<subrule><[^>|\s]+>)|([<]+)"

        # Read in BNF grammar, set production rules, terminals and
        # non-terminals.
        self.read_bnf_file(file_name)

        # Check the minimum depths of all non-terminals in the grammar.
        self.check_depths()

        # Check which non-terminals are recursive.
        self.check_recursion(self.start_rule["symbol"], [])

        # Set the minimum path and maximum arity of the grammar.
        self.set_arity()

        # Generate lists of recursive production choices and shortest
        # terminating path production choices for each NT in the grammar.
        # Enables faster tree operations.
        self.set_grammar_properties()

        # Calculate the total number of derivation tree permutations and
        # combinations that can be created by a grammar at a range of depths.
        self.check_permutations()

        if params["MIN_INIT_TREE_DEPTH"]:
            # Set the minimum ramping tree depth from the command line.
            self.min_ramp = params["MIN_INIT_TREE_DEPTH"]

        elif hasattr(params["INITIALISATION"], "ramping"):
            # Set the minimum depth at which ramping can start where we can
            # have unique solutions (no duplicates).
            self.get_min_ramp_depth()

        if params["REVERSE_MAPPING_TARGET"] or params["TARGET_SEED_FOLDER"]:
            # Initialise dicts for reverse-mapping GE individuals.
            self.concat_NTs, self.climb_NTs = {}, {}

            # Find production choices which can be used to concatenate
            # subtrees.
            self.find_concatenation_NTs()

    def read_bnf_file(self, file_name):
        """
        Read a grammar file in BNF format. Parses the grammar and saves a
        dict of all production rules and their possible choices.

        :param file_name: A specified BNF grammar file.
        :return: Nothing.
        """

        with open(file_name, "r") as bnf:
            # Read the whole grammar file.
            content = bnf.read()

            for rule in finditer(self.ruleregex, content, DOTALL):
                # Find all rules in the grammar

                if self.start_rule is None:
                    # Set the first rule found as the start rule.
                    self.start_rule = {"symbol": rule.group(
                        "rulename"), "type": "NT"}

                # Create and add a new rule.
                self.non_terminals[rule.group("rulename")] = {
                    "id": rule.group("rulename"),
                    "min_steps": maxsize,
                    "expanded": False,
                    "recursive": True,
                    "b_factor": 0,
                }

                # Initialise empty list of all production choices for this
                # rule.
                tmp_productions = []

                for p in finditer(
                    self.productionregex, rule.group("production"), MULTILINE
                ):
                    # Iterate over all production choices for this rule.
                    # Split production choices of a rule.

                    if p.group("production") is None or p.group("production").isspace():
                        # Skip to the next iteration of the loop if the
                        # current "p" production is None or blank space.
                        continue

                    # Initialise empty data structures for production choice
                    tmp_production, terminalparts = [], None

                    # special cases: GE_RANGE:dataset_n_vars will be
                    # transformed to productions 0 | 1 | ... |
                    # n_vars-1, and similar for dataset_n_is,
                    # dataset_n_os
                    GE_RANGE_regex = r"GE_RANGE:(?P<range>\w*)"
                    m = match(GE_RANGE_regex, p.group("production"))
                    if m:
                        try:
                            if m.group("range") == "dataset_n_vars":
                                # number of columns from dataset
                                n = params["FITNESS_FUNCTION"].n_vars
                            elif m.group("range") == "dataset_n_is":
                                # number of input symbols (see
                                # if_else_classifier.py)
                                n = params["FITNESS_FUNCTION"].n_is
                            elif m.group("range") == "dataset_n_os":
                                # number of output symbols
                                n = params["FITNESS_FUNCTION"].n_os
                            else:
                                # assume it's just an int
                                n = int(m.group("range"))
                        except (ValueError, AttributeError):
                            raise ValueError(
                                "Bad use of GE_RANGE: " + m.group())

                        for i in range(n):
                            # add a terminal symbol
                            tmp_production, terminalparts = [], None
                            symbol = {
                                "symbol": str(i),
                                "type": "T",
                                "min_steps": 0,
                                "recursive": False,
                            }
                            tmp_production.append(symbol)
                            if str(i) not in self.terminals:
                                self.terminals[str(i)] = [
                                    rule.group("rulename")]
                            elif rule.group("rulename") not in self.terminals[str(i)]:
                                self.terminals[str(i)].append(
                                    rule.group("rulename"))
                            tmp_productions.append(
                                {
                                    "choice": tmp_production,
                                    "recursive": False,
                                    "NT_kids": False,
                                }
                            )
                        # don't try to process this production further
                        # (but later productions in same rule will work)
                        continue

                    for sub_p in finditer(
                        self.productionpartsregex, p.group(
                            "production").strip()
                    ):
                        # Split production into terminal and non terminal
                        # symbols.

                        if sub_p.group("subrule"):
                            if terminalparts is not None:
                                # Terminal symbol is to be appended to the
                                # terminals dictionary.
                                symbol = {
                                    "symbol": terminalparts,
                                    "type": "T",
                                    "min_steps": 0,
                                    "recursive": False,
                                }
                                tmp_production.append(symbol)
                                if terminalparts not in self.terminals:
                                    self.terminals[terminalparts] = [
                                        rule.group("rulename")
                                    ]
                                elif (
                                    rule.group("rulename")
                                    not in self.terminals[terminalparts]
                                ):
                                    self.terminals[terminalparts].append(
                                        rule.group("rulename")
                                    )
                                terminalparts = None

                            tmp_production.append(
                                {"symbol": sub_p.group(
                                    "subrule"), "type": "NT"}
                            )

                        else:
                            # Unescape special characters (\n, \t etc.)
                            if terminalparts is None:
                                terminalparts = ""
                            terminalparts += "".join(
                                [
                                    part.encode().decode("unicode-escape")
                                    for part in sub_p.groups()
                                    if part
                                ]
                            )

                    if terminalparts is not None:
                        # Terminal symbol is to be appended to the terminals
                        # dictionary.
                        symbol = {
                            "symbol": terminalparts,
                            "type": "T",
                            "min_steps": 0,
                            "recursive": False,
                        }
                        tmp_production.append(symbol)
                        if terminalparts not in self.terminals:
                            self.terminals[terminalparts] = [
                                rule.group("rulename")]
                        elif (
                            rule.group(
                                "rulename") not in self.terminals[terminalparts]
                        ):
                            self.terminals[terminalparts].append(
                                rule.group("rulename"))
                    tmp_productions.append(
                        {"choice": tmp_production,
                            "recursive": False, "NT_kids": False}
                    )

                if not rule.group("rulename") in self.rules:
                    # Add new production rule to the rules dictionary if not
                    # already there.
                    self.rules[rule.group("rulename")] = {
                        "choices": tmp_productions,
                        "no_choices": len(tmp_productions),
                    }

                    if len(tmp_productions) == 1:
                        # Unit productions.
                        print(
                            "Warning: Grammar contains unit production "
                            "for production rule",
                            rule.group("rulename"),
                        )
                        print("         Unit productions consume GE codons.")
                else:
                    # Conflicting rules with the same name.
                    raise ValueError("lhs should be unique",
                                     rule.group("rulename"))

    def check_depths(self):
        """
        Run through a grammar and find out the minimum distance from each
        NT to the nearest T. Useful for initialisation methods where we
        need to know how far away we are from fully expanding a tree
        relative to where we are in the tree and what the depth limit is.

        :return: Nothing.
        """

        # Initialise graph and counter for checking minimum steps to Ts for
        # each NT.
        counter, graph = 1, []

        for rule in sorted(self.rules.keys()):
            # Iterate over all NTs.
            choices = self.rules[rule]["choices"]

            # Set branching factor for each NT.
            self.non_terminals[rule]["b_factor"] = self.rules[rule]["no_choices"]

            for choice in choices:
                # Add a new edge to our graph list.
                graph.append([rule, choice["choice"]])

        while graph:
            removeset = set()
            for edge in graph:
                # Find edges which either connect to terminals or nodes
                # which are fully expanded.
                if all(
                    [
                        sy["type"] == "T"
                        or self.non_terminals[sy["symbol"]]["expanded"]
                        for sy in edge[1]
                    ]
                ):
                    removeset.add(edge[0])

            for s in removeset:
                # These NTs are now expanded and have their correct minimum
                # path set.
                self.non_terminals[s]["expanded"] = True
                self.non_terminals[s]["min_steps"] = counter

            # Create new graph list and increment counter.
            graph = [e for e in graph if e[0] not in removeset]
            counter += 1

    def check_recursion(self, cur_symbol, seen):
        """
        Traverses the grammar recursively and sets the properties of each rule.

        :param cur_symbol: symbol to check.
        :param seen: Contains already checked symbols in the current traversal.
        :return: Boolean stating whether or not cur_symbol is recursive.
        """

        if cur_symbol not in self.non_terminals.keys():
            # Current symbol is a T.
            return False

        if cur_symbol in seen:
            # Current symbol has already been seen, is recursive.
            return True

        # Append current symbol to seen list.
        seen.append(cur_symbol)

        # Get choices of current symbol.
        choices = self.rules[cur_symbol]["choices"]
        nt = self.non_terminals[cur_symbol]

        recursive = False
        for choice in choices:
            for sym in choice["choice"]:
                # Recurse over choices.
                recursive_symbol = self.check_recursion(sym["symbol"], seen)
                recursive = recursive or recursive_symbol

        # Set recursive properties.
        nt["recursive"] = recursive
        seen.remove(cur_symbol)

        return nt["recursive"]

    def set_arity(self):
        """
        Set the minimum path of the grammar, i.e. the smallest legal
        solution that can be generated.

        Set the maximum arity of the grammar, i.e. the longest path to a
        terminal from any non-terminal.

        :return: Nothing
        """

        # Set the minimum path of the grammar as the minimum steps to a
        # terminal from the start rule.
        self.min_path = self.non_terminals[self.start_rule["symbol"]]["min_steps"]

        # Set the maximum arity of the grammar as the longest path to
        # a T from any NT.
        self.max_arity = max(
            self.non_terminals[NT]["min_steps"] for NT in self.non_terminals
        )

        # Add the minimum terminal path to each production rule.
        for rule in self.rules:
            for choice in self.rules[rule]["choices"]:
                NT_kids = [i for i in choice["choice"] if i["type"] == "NT"]
                if NT_kids:
                    choice["NT_kids"] = True
                    for sym in NT_kids:
                        sym["min_steps"] = self.non_terminals[sym["symbol"]][
                            "min_steps"
                        ]

        # Add boolean flag indicating recursion to each production rule.
        for rule in self.rules:
            for prod in self.rules[rule]["choices"]:
                for sym in [i for i in prod["choice"] if i["type"] == "NT"]:
                    sym["recursive"] = self.non_terminals[sym["symbol"]]["recursive"]
                    if sym["recursive"]:
                        prod["recursive"] = True

    def set_grammar_properties(self):
        """
        Goes through all non-terminals and finds the production choices with
        the minimum steps to terminals and with recursive steps.

        :return: Nothing
        """

        for nt in self.non_terminals:
            # Loop over all non terminals.
            # Find the production choices for the current NT.
            choices = self.rules[nt]["choices"]

            for choice in choices:
                # Set the maximum path to a terminal for each produciton choice
                choice["max_path"] = max(
                    [item["min_steps"] for item in choice["choice"]]
                )

            # Find shortest path to a terminal for all production choices for
            # the current NT. The shortest path will be the minimum of the
            # maximum paths to a T for each choice over all chocies.
            min_path = min([choice["max_path"] for choice in choices])

            # Set the minimum path in the self.non_terminals dict.
            self.non_terminals[nt]["min_path"] = [
                choice for choice in choices if choice["max_path"] == min_path
            ]

            # Find recursive production choices for current NT. If any
            # constituent part of a production choice is recursive,
            # it is added to the recursive list.
            self.non_terminals[nt]["recursive"] = [
                choice for choice in choices if choice["recursive"]
            ]

    def check_permutations(self):
        """
        Calculates how many possible derivation tree combinations can be
        created from the given grammar at a specified depth. Only returns
        possible combinations at the specific given depth (if there are no
        possible permutations for a given depth, will return 0).

        :param ramps:
        :return: Nothing.
        """

        # Set the number of depths permutations are calculated for
        # (starting from the minimum path of the grammar)
        ramps = params["PERMUTATION_RAMPS"]

        perms_list = []
        if self.max_arity > self.min_path:
            for i in range(max((self.max_arity + 1 - self.min_path), ramps)):
                x = self.check_all_permutations(i + self.min_path)
                perms_list.append(x)
                if i > 0:
                    perms_list[i] -= sum(perms_list[:i])
                    self.permutations[i + self.min_path] -= sum(perms_list[:i])
        else:
            for i in range(ramps):
                x = self.check_all_permutations(i + self.min_path)
                perms_list.append(x)
                if i > 0:
                    perms_list[i] -= sum(perms_list[:i])
                    self.permutations[i + self.min_path] -= sum(perms_list[:i])

    def check_all_permutations(self, depth):
        """
        Calculates how many possible derivation tree combinations can be
        created from the given grammar at a specified depth. Returns all
        possible combinations at the specific given depth including those
        depths below the given depth.

        :param depth: A depth for which to calculate the number of
        permutations of solution that can be generated by the grammar.
        :return: The permutations possible at the given depth.
        """

        if depth < self.min_path:
            # There is a bug somewhere that is looking for a tree smaller than
            # any we can create
            s = (
                "representation.grammar.Grammar.check_all_permutations\n"
                "Error: cannot check permutations for tree smaller than the "
                "minimum size."
            )
            raise Exception(s)

        if depth in self.permutations.keys():
            # We have already calculated the permutations at the requested
            # depth.
            return self.permutations[depth]

        else:
            # Calculate permutations at the requested depth.
            # Initialise empty data arrays.
            pos, depth_per_symbol_trees, productions = 0, {}, []

            for NT in self.non_terminals:
                # Iterate over all non-terminals to fill out list of
                # productions which contain non-terminal choices.
                a = self.non_terminals[NT]

                for rule in self.rules[a["id"]]["choices"]:
                    if rule["NT_kids"]:
                        productions.append(rule)

            # Get list of all production choices from the start symbol.
            start_symbols = self.rules[self.start_rule["symbol"]]["choices"]

            for choice in productions:
                # Generate a list of the symbols of each production choice
                key = str([sym["symbol"] for sym in choice["choice"]])

                # Initialise permutations dictionary with the list
                depth_per_symbol_trees[key] = {}

            for i in range(2, depth + 1):
                # Find all the possible permutations from depth of min_path up
                # to a specified depth

                for choice in productions:
                    # Iterate over all production choices
                    sym_pos = 1

                    for j in choice["choice"]:
                        # Iterate over all symbols in a production choice.
                        symbol_arity_pos = 0

                        if j["type"] is "NT":
                            # We are only interested in non-terminal symbols
                            for child in self.rules[j["symbol"]]["choices"]:
                                # Iterate over all production choices for
                                # each NT symbol in the original choice.

                                if (
                                    len(child["choice"]) == 1
                                    and child["choice"][0]["type"] == "T"
                                ):
                                    # If the child choice leads directly to
                                    # a single terminal, increment the
                                    # permutation count.
                                    symbol_arity_pos += 1

                                else:
                                    # The child choice does not lead
                                    # directly to a single terminal.
                                    # Generate a key for the permutations
                                    # dictionary and increment the
                                    # permutations count there.
                                    key = [sym["symbol"]
                                           for sym in child["choice"]]
                                    if (i - 1) in depth_per_symbol_trees[
                                        str(key)
                                    ].keys():
                                        symbol_arity_pos += depth_per_symbol_trees[
                                            str(key)
                                        ][i - 1]

                            # Multiply original count by new count.
                            sym_pos *= symbol_arity_pos

                    # Generate new key for the current production choice and
                    # set the new value in the permutations dictionary.
                    key = [sym["symbol"] for sym in choice["choice"]]
                    depth_per_symbol_trees[str(key)][i] = sym_pos

            # Calculate permutations for the start symbol.
            for sy in start_symbols:
                key = [sym["symbol"] for sym in sy["choice"]]
                if str(key) in depth_per_symbol_trees:
                    pos += (
                        depth_per_symbol_trees[str(key)][depth]
                        if depth in depth_per_symbol_trees[str(key)]
                        else 0
                    )
                else:
                    pos += 1

            # Set the overall permutations dictionary for the current depth.
            self.permutations[depth] = pos

            return pos

    def get_min_ramp_depth(self):
        """
        Find the minimum depth at which ramping can start where we can have
        unique solutions (no duplicates).

        :param self: An instance of the representation.grammar.grammar class.
        :return: The minimum depth at which unique solutions can be generated
        """

        max_tree_depth = params["MAX_INIT_TREE_DEPTH"]
        size = params["POPULATION_SIZE"]

        # Specify the range of ramping depths
        depths = range(self.min_path, max_tree_depth + 1)

        if size % 2:
            # Population size is odd
            size += 1

        if size / 2 < len(depths):
            # The population size is too small to fully cover all ramping
            # depths. Only ramp to the number of depths we can reach.
            depths = depths[: int(size / 2)]

        # Find the minimum number of unique solutions required to generate
        # sufficient individuals at each depth.
        unique_start = int(floor(size / len(depths)))
        ramp = None

        for i in sorted(self.permutations.keys()):
            # Examine the number of permutations and combinations of unique
            # solutions capable of being generated by a grammar across each
            # depth i.
            if self.permutations[i] > unique_start:
                # If the number of permutations possible at a given depth i is
                # greater than the required number of unique solutions,
                # set the minimum ramp depth and break out of the loop.
                ramp = i
                break
        self.min_ramp = ramp

    def find_concatenation_NTs(self):
        """
        Scour the grammar class to find non-terminals which can be used to
        combine/reduce_trees derivation trees. Build up a list of such
        non-terminals. A concatenation non-terminal is one in which at least
        one production choice contains multiple non-terminals. For example:

            <e> ::= (<e><o><e>)|<v>

        is a concatenation NT, since the production choice (<e><o><e>) can
        reduce_trees multiple NTs together. Note that this choice also includes
        a combination of terminals and non-terminals.

        :return: Nothing.
        """

        # Iterate over all non-terminals/production rules.
        for rule in sorted(self.rules.keys()):

            # Find rules which have production choices leading to NTs.
            concat = [
                choice for choice in self.rules[rule]["choices"] if choice["NT_kids"]
            ]

            if concat:
                # We can reduce_trees NTs.
                for choice in concat:

                    symbols = [[sym["symbol"], sym["type"]]
                               for sym in choice["choice"]]

                    NTs = [
                        sym["symbol"] for sym in choice["choice"] if sym["type"] == "NT"
                    ]

                    for NT in NTs:
                        # We add to our self.concat_NTs dictionary. The key is
                        # the root node we want to reduce_trees with another
                        # node. This way when we have a node and wish to see
                        # if we can reduce_trees it with anything else, we
                        # simply look up this dictionary.
                        conc = [choice["choice"], rule, symbols]

                        if NT not in self.concat_NTs:
                            self.concat_NTs[NT] = [conc]
                        else:
                            if conc not in self.concat_NTs[NT]:
                                self.concat_NTs[NT].append(conc)

    def __str__(self):
        return "%s %s %s %s" % (
            self.terminals,
            self.non_terminals,
            self.rules,
            self.start_rule,
        )



#!/usr/bin/env python
"""
@author: Jonathan Byrne
@17/01/18 11:09
"""

import json
import time
from os import path

import numpy as np
# imports relative to keras model (ind.phenotype)
from tensorflow.keras import Input, Sequential
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.layers import BatchNormalization, Dense, Dropout
from tensorflow.keras.utils import plot_model

from autooc.algorithm.parameters import params
from autooc.fitness.base_ff_classes.base_ff import base_ff
from autooc.stats.stats import stats
from autooc.utilities.utils import get_model_from_encoder


class singlefit_autoencoders(base_ff):
    """
    An example of a single fitness class that generates
    two fitness values for multi-objective optimisation
    """

    maximise = True
    multi_objective = True

    def __init__(self):

        # Initialise base fitness function class.
        super().__init__()

        # Set list of individual fitness functions.
        self.num_obj = 2
        dummyfit = base_ff()
        dummyfit.maximise = True
        self.fitness_functions = [dummyfit, dummyfit]
        self.default_fitness = [float("nan"), float("nan")]

        self.dtrain, self.dval = params["normal_train_data"], params["test_data"]

    def evaluate(self, ind, **kwargs):
        """Dummy fitness function that generates 2 fitness values"""
        phenotype = ind.phenotype
        fitness = 0
        settings = {}

        d = {
            "Sequential": Sequential,
            "Dense": Dense,
            "BatchNormalization": BatchNormalization,
            "Input": Input,
            "Dropout": Dropout,
            "get_model_from_encoder": get_model_from_encoder,
        }

        exec(ind.phenotype, d)
        print("FENOTIPO")
        print(ind.phenotype)
        print("END FENOTIPO")
        # print(type(ind.phenotype))
        print("\nSTART D:\n")
        print(d)
        print("\nEND D:\n")

        model = d["model"]

        rand_name = str(stats["gen"]) + "_" + str(ind.name) + "_model.png"
        filename = path.join(params["FILE_PATH"], rand_name)
        plot_model(model, filename, show_shapes=True)
        early_stop = EarlyStopping(
            monitor="val_loss",
            patience=10,
            verbose=1,
            mode="min",
            restore_best_weights=True,
        )
        start = time.time()
        hist = model.fit(
            self.dtrain,
            self.dtrain,
            epochs=params["EPOCHS"],
            validation_data=(self.dval, self.dval),
            verbose=1,
            callbacks=[early_stop],
        )
        end = time.time()
        # params["TRAINING_TIME"] = end-start
        training_time = start - end
        rand_name = str(stats["gen"]) + "_" + str(ind.name) + "_hist.json"
        filename = path.join(params["FILE_PATH"], rand_name)
        with open(filename, "w") as f:
            json.dump(hist.history, f)

        auc = params["ERROR_METRIC"](model)

        filename = path.join(params["FILE_PATH"], "test_aucs.csv")
        with open(filename, "a") as the_file:
            the_file.write(
                (str(stats["gen"]) + ";" +
                 str(ind.name) + ";" + str(auc) + "\n")
            )
        print(auc)

        fitness = [auc, training_time]

        return fitness

    @staticmethod
    def value(fitness_vector, objective_index):
        """
        This is a static method required by NSGA-II for sorting populations
        based on a given fitness function, or for returning a given index of a
        population based on a given fitness function.

        :param fitness_vector: A vector/list of fitnesses.
        :param objective_index: The index of the desired fitness.
        :return: The fitness at the objective index of the fitness vector.
        """

        if not isinstance(fitness_vector, list):
            return float("inf")

        return fitness_vector[objective_index]



import os
import sys
from sys import path

import utilities.algorithm.command_line_parser as parser

from autooc.algorithm.algorithm.general import check_python_version
from autooc.algorithm.fitness.math_functions import sci_notation
from autooc.algorithm.parameters import params
from autooc.representation.grammar import Grammar

path.append("../src")


check_python_version()


def main(command_line_args):
    """
    Given a specified grammar file, parse the grammar using the Grammar class
    and print out the number of unique permutations and combinations of
    distinct phenotypes that this grammar is capable of generating at a
    number of given depths.

    :return: Nothing.
    """

    # Parse command line args (we only want the grammar file)
    cmd_args, unknown = parser.parse_cmd_args(command_line_args)

    # Join original params dictionary with command line specified arguments.
    # NOTE that command line arguments overwrite all previously set parameters.
    params.update(cmd_args)

    # Parse grammar file and set grammar class.
    grammar = Grammar(os.path.join("..", "grammars", params["GRAMMAR_FILE"]))

    print("\nSpecified grammar:", params["GRAMMAR_FILE"])

    # Initialise zero maximum branching factor for grammar
    max_b_factor = 0

    print("\nBranching factor for each non-terminal:")

    for NT in sorted(grammar.non_terminals.keys()):

        # Get branching factor for current NT.
        b_factor = grammar.non_terminals[NT]["b_factor"]

        # Print info.
        print("", NT, "   \t:", b_factor)

        # Set maximum branching factor.
        if b_factor > max_b_factor:
            max_b_factor = b_factor

    print("\nMaximum branching factor of the grammar:", max_b_factor)

    # Initialise counter for the total number of solutions.
    total_solutions = 0

    print("\nNumber of unique possible solutions for a range of depths:\n")

    for depth in grammar.permutations:

        # Get number of solutions possible at current depth
        solutions = grammar.permutations[depth]

        print(
            " Depth: %d \t Number of unique solutions: %s"
            % (depth, sci_notation(solutions))
        )

        # Increment total number of solutions.
        total_solutions += solutions

    print(
        "\nTotal number of unique possible solutions that can be generated"
        "up to and including a depth of %d: %s" % (
            depth, sci_notation(total_solutions))
    )


if __name__ == "__main__":

    # Do not write or save any files.
    params["DEBUG"] = True

    # Run main program.
    main(sys.argv[1:])  # exclude the ponyge.py arg itself



from autooc.algorithm.fitness.error_metric import calculate_bic
from autooc.algorithm.parameters import params
from autooc.fitness.base_ff_classes.base_ff import base_ff


class bic(base_ff):
    """
    Fitness function class for calculating Bayesian Information Criterion
    """

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

    def evaluate(self, ind, **kwargs):
        return calculate_bic(ind.model, params["X_train_len"])



"""Latent-Tree GE (LTGE)

Alberto Moraglio and James McDermott 2018.

This is still GE, with integer genomes, but integers arranged in a
dictionary. In the dictionary of key-value pairs, each key is a list
of tuples representing a path from the root to the node, and the
corresponding value is an integer which (still) represents the choice
of production made at the node. We call it latent tree GE because the
representation itself is "flat" (a dictionary) but the derivation tree
is implicit or latent.

Example genome with interpretation:

{
(): 8, # at the root, choice 8 was made
((8, 1),): 5, # for the first symbol in choice 8, choice 5 was made
((8, 1), (5, 1)): 4,
((8, 1), (5, 1), (4, 1)): 9
}


The initialisation operator in LTGE is equivalent to that in GE.
The mutation operator is equivalent to the CFG-GP mutation.  The
crossover operator is equivalent to a homologous crossover on
derivation trees.

"""


import random

from autooc.algorithm.parameters import params
from autooc.representation.derivation import legal_productions


def latent_tree_random_ind(grammar, maxdepth, old_genome=None):
    """Generate a random individual (genome and string), OR repair a
    genome and generate its string. These two things are conceptually
    distinct, but they share almost all code. The idea is that
    repairing a genome consists of traversing it and using its choices
    where they are appropriate, and otherwise generating new choices
    (and any old, unused choices will be discarded). Generating a new
    individual just means that there are no appropriate choices, so
    always generating new choices. So, we implement both in the same
    function."""

    # we use this inner helper function to do the work.
    # it "accumulates" the genome as it runs.
    def _random_ind(gram, genome, depth, s=None, name=None):
        """Recursively create a genome. gram is a grammar, genome a dict
        (initially empty), depth an integer giving maximum depth. s is
        the current symbol (None tells us to use the start symbol.) name
        is the name-in-progress."""
        if s is None:
            s = gram.start_rule["symbol"]
            name = tuple()
        elif s in gram.terminals:
            return s

        rule = gram.rules[s]

        if old_genome and name in old_genome:

            # A valid entry was found in old_genome. Apply mod rule as
            # part of repair: it will have no effect if the value is
            # already in the right range.
            gi = old_genome[name] % len(rule["choices"])
            prod = rule["choices"][gi]

        else:

            # No valid entry was found, so choose a production from
            # among those which are legal (because their min depth to
            # finish recursion is less than or equal to max depth
            # minus our current depth).
            productions = params["BNF_GRAMMAR"].rules[s]
            available = legal_productions(
                "random", depth, s, productions["choices"])
            prod = random.choice(available)  # choose production
            gi = productions["choices"].index(prod)  # find its index

        genome[name] = gi

        # Join together all the strings for this production. For
        # terminals just use the string itself. For non-terminals,
        # recurse: decrease depth, pass in the symbol, and append to
        # the naming scheme according to the choice we made in the
        # current call. Recall that each symbol s is a dict:
        # s["symbol"] is the symbol itself, s["type"] is 'T' or 'NT'.
        return "".join(
            (
                s["symbol"]
                if s["type"] == "T"
                else _random_ind(
                    gram, genome, depth - 1, s["symbol"], name + ((gi, i),)
                )
            )
            for i, s in enumerate(prod["choice"])
        )

    genome = {}
    s = _random_ind(grammar, genome, maxdepth, None, None)
    return genome, s


def latent_tree_repair(genome, gram, maxdepth):
    """Given a genome, make any necessary repairs. This could include
    discarding unused genetic material, generating new material, and
    taking the 'mod' of existing values. It is just a wrapper on
    random_ind which does all the work. It re-orders the arguments
    since the object of the verb repair is the thing to be repaired
    (the genome) whereas in random_ind the genome is an optional final
    argument."""
    return latent_tree_random_ind(gram, maxdepth, genome)


def latent_tree_crossover(g1, g2):
    """Produce a single child genome by crossover through dict
    manipulation. For each key: if present in both parents, then
    choose value randomly; if present in only one parent, then use
    that. Later, repair must make sure the offspring is valid."""

    # Put all of g1 in, to start with. FIXME The deep_copy() in
    # crossover_inds() ought to guarantee us that g1 and g2 are
    # copies, not original members of the population so we can edit
    # them in-place. However, I'm finding that c = g1 is giving
    # different results from c = g1.copy(). I've checked that
    # deep_copy() is actually running, so I don't understand this
    # problem. To be on the safe side, we'll keep the copy() here.
    # See https://github.com/PonyGE/PonyGE2/issues/89.
    c = g1.copy()
    for k in g2.keys():
        if k in g1:
            # k is in both parents so choose randomly.
            c[k] = random.choice((g1[k], g2[k]))
        else:
            # for items in g2 only, copy them in
            c[k] = g2[k]
    return c


def latent_tree_mutate(g):
    """Produce an offspring genome by mutation through dict
    manipulation. Choose a random key in the dict, and overwrite its
    value with a random int. Later, repair must make sure the
    offspring is valid, including using the mod rule to map from a
    (possibly) large int to the corresponding small one (ie the one
    giving the same production choice) in the range of possible
    choices."""

    # FIXME We don't rely on g being a copy, in case the search
    # algorithm sometimes mutates individuals which are original
    # members of the population.
    # See https://github.com/PonyGE/PonyGE2/issues/89.
    g = g.copy()
    k = random.choice(list(g.keys()))
    g[k] = random.randrange(1000000)  # there is no true maxint on py 3
    return g



import itertools
import random

import numpy as np

from autooc.algorithm.parameters import params
from autooc.fitness.supervised_learning.supervised_learning import \
    supervised_learning
from autooc.utilities.fitness.error_metric import Hamming_error


class if_else_classifier(supervised_learning):
    """Fitness function for if-else classifier problems. We
    specialise the supervised learning fitness function.

    The user must pass in n_vars, n_is, n_os (number of variables,
    input symbols, output symbols). This is accomplished with, eg,
    --extra_parameters 3 10 2.

    The input and output symbols are just integers. The target
    function is:

    ((x[0] + x[1]) % n_os) + 1

    The candidate solutions are like this:

    (3 if x[0] == 2 else (5 if (x[1] == 1 and x[2] == 1) else 9))

    The possible outputs (3, 5, 9) are in the range (0, n_os-1).  The
    inputs (2, 1) are in the range (0, n_is-1). The variables x[i] are
    from x[0] to x[n_vars-1].

    The target function is evaluated at all possible inputs. That
    gives a target dataset for training. There is no test on unseen
    data.

    An example command-line is then:

    python ponyge.py --generations 10 --population 10 --fitness
    supervised_learning.if_else_classifier
    --extra_parameters 3 10 2 --grammar
    supervised_learning/if_else_classifier.bnf

    """

    def __init__(self):
        # Don't call super().__init__() because it reads the training
        # (and test) data from files. We'll do everything else it
        # would have done.

        # we created n, n_is, n_os for convenience, but also the
        # variables with full names hanging on self since grammar.py
        # may need them for GE_RANGE:dataset_n_vars etc.
        n = self.n_vars = int(params["EXTRA_PARAMETERS"][0])
        n_is = self.n_is = int(params["EXTRA_PARAMETERS"][1])
        n_os = self.n_os = int(params["EXTRA_PARAMETERS"][2])

        # Set error metric if it's not set already.
        if params["ERROR_METRIC"] is None:
            params["ERROR_METRIC"] = Hamming_error

        self.maximise = params["ERROR_METRIC"].maximise

        # create the target function
        target = target_classifier(n, n_is, n_os)

        # generate all input combinations for n variables, to become
        # the fitness cases
        Ls = [list(range(n_is)) for i in range(n)]
        X = np.array(list(itertools.product(*Ls)))
        # evaluate the target function at the fitness cases
        self.training_exp = np.array([target(xi) for xi in X])
        self.training_in = X.T

        # In these Classifier problems we don't want a separate test
        # set, and we don't optimize constants.
        assert not params["DATASET_TEST"]
        assert not params["OPTIMIZE_CONSTANTS"]


def target_classifier(n_vars, n_is, n_os):
    def target(x):
        return ((x[0] + x[1]) % n_os) + 1

    return target



from autooc.fitness.base_ff_classes.base_ff import base_ff
from autooc.utilities.fitness.math_functions import binary_phen_to_float


class binary_phenotype_to_float(base_ff):

    """
    Fitness function for the first problem (T_1) presented in
    [Zitzler2000].

    .. Zitzler, Eckart, Kalyanmoy Deb, and Lothar Thiele. Comparison
    of multiobjective evolutionary algorithms: Empirical results.
    Evolutionary computation 8.2 (2000): 173-195.
    """

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

    def evaluate(self, ind, **kwargs):

        min_value = [0] * 30
        max_value = [1] * 30

        real_chromosome = binary_phen_to_float(
            ind.phenotype, 30, min_value, max_value)

        return real_chromosome[0]



import warnings
from math import log

import numpy as np
import tensorflow as tf
from sklearn.metrics import auc
from sklearn.metrics import f1_score as sklearn_f1_score
from sklearn.metrics import log_loss, roc_auc_score
from tqdm import tqdm

from autooc.algorithm.parameters import params


def mae(y, yhat):
    """
    Calculate mean absolute error between inputs.

    :param y: The expected input (i.e. from dataset).
    :param yhat: The given input (i.e. from phenotype).
    :return: The mean absolute error.
    """

    return np.mean(np.abs(y - yhat))


# Set maximise attribute for mae error metric.
mae.maximise = False


def rmse(y, yhat):
    """
    Calculate root mean square error between inputs.

    :param y: The expected input (i.e. from dataset).
    :param yhat: The given input (i.e. from phenotype).
    :return: The root mean square error.
    """

    return np.sqrt(np.mean(np.square(y - yhat)))


# Set maximise attribute for rmse error metric.
rmse.maximise = False


def mse(y, yhat):
    """
    Calculate mean square error between inputs.

    :param y: The expected input (i.e. from dataset).
    :param yhat: The given input (i.e. from phenotype).
    :return: The mean square error.
    """

    return np.mean(np.square(y - yhat))


# Set maximise attribute for mse error metric.
mse.maximise = False


def hinge(y, yhat):
    """
    Hinge loss is a suitable loss function for classification.  Here y is
    the true values (-1 and 1) and yhat is the "raw" output of the individual,
    ie a real value. The classifier will use sign(yhat) as its prediction.

    :param y: The expected input (i.e. from dataset).
    :param yhat: The given input (i.e. from phenotype).
    :return: The hinge loss.
    """

    # Deal with possibility of {-1, 1} or {0, 1} class label convention
    y_vals = set(y)
    # convert from {0, 1} to {-1, 1}
    if 0 in y_vals:
        y[y == 0] = -1

    # Our definition of hinge loss cannot be used for multi-class
    assert len(y_vals) == 2

    # NB not np.max. maximum does element-wise max.  Also we use the
    # mean hinge loss rather than sum so that the result doesn't
    # depend on the size of the dataset.
    return np.mean(np.maximum(0, 1 - y * yhat))


# Set maximise attribute for hinge error metric.
hinge.maximise = False


def f1_score(y, yhat):
    """
    The F_1 score is a metric for classification which tries to balance
    precision and recall, ie both true positives and true negatives.
    For F_1 score higher is better.

    :param y: The expected input (i.e. from dataset).
    :param yhat: The given input (i.e. from phenotype).
    :return: The f1 score.
    """

    # if phen is a constant, eg 0.001 (doesn't refer to x), then yhat
    # will be a constant. that will break f1_score. so convert to a
    # constant array.
    if not isinstance(yhat, np.ndarray) or len(yhat.shape) < 1:
        yhat = np.ones_like(y) * yhat

    # Deal with possibility of {-1, 1} or {0, 1} class label
    # convention.  FIXME: would it be better to canonicalise the
    # convention elsewhere and/or create user parameter to control it?
    # See https://github.com/PonyGE/PonyGE2/issues/113.
    y_vals = set(y)
    # convert from {-1, 1} to {0, 1}
    if -1 in y_vals:
        y[y == -1] = 0

    # We binarize with a threshold, so this cannot be used for multi-class
    assert len(y_vals) == 2

    # convert real values to boolean {0, 1} with a zero threshold
    yhat = yhat > 0

    with warnings.catch_warnings():
        # if we predict the same value for all samples (trivial
        # individuals will do so as described above) then f-score is
        # undefined, and sklearn will give a runtime warning and
        # return 0. We can ignore that warning and happily return 0.
        warnings.simplefilter("ignore")
        return sklearn_f1_score(y, yhat, average="weighted")


# Set maximise attribute for f1_score error metric.
f1_score.maximise = True


def AUC(y, yhat):
    """
    Calculate the Area Under the Curve (AUC).
    :param y: The expected input (i.e. from dataset).
    :param yhat: The given input (i.e. from phenotype).
    :return: The AUC.
    """
    ### NEW ###
    # Replaces NaNs or infinite values
    yhat = np.nan_to_num(yhat)
    if type(yhat) != np.ndarray:
        yhat = np.repeat(yhat, len(y))
    # auc_val = roc_auc_score(y, yhat)*100
    auc_score = roc_auc_score(y, yhat)
    # auc_val = auc(fpr, tpr) * 100
    return auc_score

    # print(auc)
    # return(auc)


# Set maximise attribute for auc error metric.
AUC.maximise = True

"""def predict(model, data, threshold):
    reconstructions = model(data)
    loss = tf.keras.losses.mae(reconstructions, data)
    return tf.math.less(loss, threshold)

def print_stats(predictions, labels):
    print("Accuracy = {}".format(accuracy_score(labels, predictions)))
    print("Precision = {}".format(precision_score(labels, predictions)))
    print("Recall = {}".format(recall_score(labels, predictions)))
    print("AUC = {}".format(roc_auc_score(labels, predictions)))

preds = predict(autoencoder, test_data, threshold)
print_stats(preds, test_labels)"""


def auc_autoencoder(model):
    rec_error, rec_losses, threshold = reconstruction_error(model)

    X_val = params["X_val"]
    y_val = params["y_val"]

    #predictions = tf.math.less(rec_losses, threshold)
    predictions = [(lambda x: params["ANOMALY_CLASS"] if x >
                    threshold else params["NORMAL_CLASS"])(x) for x in rec_losses]

    try:
        auc = roc_auc_score(y_val, predictions)
    except Exception:
        auc = 0
    # print("AUCCCCCC")
    print(auc)

    return auc, rec_error, rec_losses, threshold


auc_autoencoder.maximise = True


def auc_sklearn(model):
    X_val = params["X_val"]
    y_val = params["y_val"]

    predictions = model.predict(X_val)
    predictions = [(lambda x: params["ANOMALY_CLASS"] if x == -
                    1 else params["NORMAL_CLASS"])(x) for x in predictions]

    print(predictions)
    print(y_val)

    try:
        auc = roc_auc_score(y_val, predictions)
    except Exception:
        auc = 0
    print("AUCCCCCC")
    print(auc)

    return auc


auc_sklearn.maximise = True


def reconstruction_error(model):
    test_data = params["X_val"]
    print(test_data.shape)

    reconstructions = model.predict(test_data)
    print(reconstructions.shape)
    rec_losses = tf.keras.losses.mae(reconstructions, test_data)

    threshold = np.mean(rec_losses) + np.std(rec_losses)

    rec_error = tf.math.reduce_mean(rec_losses).numpy()
    # threshold = 0.018963557
    # predictions = tf.math.less(test_loss, threshold)

    # test_labels = params["test_labels"]
    # auc = roc_auc_score(test_labels, predictions)

    return rec_error, rec_losses, threshold


reconstruction_error.maximise = False


def anomaly_score_sklearn(model):
    test_data = params["X_val"]
    print(test_data.shape)

    anomaly_scores = model.decision_function(test_data)
    print(anomaly_scores.shape)

    return np.mean(anomaly_scores)


anomaly_score_sklearn.maximise = False


def anomaly_score_iforest(model):
    test_data = params["X_val"]
    print(test_data.shape)

    anomaly_scores = model.decision_function(test_data)
    print(anomaly_scores.shape)

    return np.mean(anomaly_scores)


anomaly_score_iforest.maximise = False


def anomaly_score_svm(model):
    test_data = params["X_val"]
    print(test_data.shape)

    anomaly_scores = model.decision_function(test_data)
    print(anomaly_scores.shape)

    return np.mean(anomaly_scores)


anomaly_score_svm.maximise = True


def anomaly_score_lof(model):
    test_data = params["X_val"]
    print(test_data.shape)

    anomaly_scores = model.decision_function(test_data)
    print(anomaly_scores.shape)

    return np.mean(anomaly_scores)


anomaly_score_lof.maximise = True


# calculate Bayesian Information Criterion
def calculate_bic(model, n):
    rec_error, _ = reconstruction_error(model)
    return n * log(rec_error) + model.count_params() * log(n)


calculate_bic.maximise = False


def Hamming_error(y, yhat):
    """
    The number of mismatches between y and yhat. Suitable
    for Boolean problems and if-else classifier problems.
    Assumes both y and yhat are binary or integer-valued.
    """
    return np.sum(y != yhat)


Hamming_error.maximise = False



from copy import copy
from sys import stdout
from time import time

import numpy as np

from autooc.algorithm.parameters import params
from autooc.utilities.algorithm.NSGA2 import compute_pareto_metrics
from autooc.utilities.algorithm.state import create_state
from autooc.utilities.stats import trackers
from autooc.utilities.stats.file_io import (save_best_ind_to_file,
                                           save_first_front_to_file,
                                           save_stats_headers,
                                           save_stats_to_file)
from autooc.utilities.stats.save_plots import (save_pareto_fitness_plot,
                                              save_plot_from_data)

"""Algorithm statistics"""
stats = {
    "gen": 0,
    "total_inds": 0,
    "regens": 0,
    "invalids": 0,
    "runtime_error": 0,
    "unique_inds": len(trackers.cache),
    "unused_search": 0,
    "ave_genome_length": 0,
    "max_genome_length": 0,
    "min_genome_length": 0,
    "ave_used_codons": 0,
    "max_used_codons": 0,
    "min_used_codons": 0,
    "ave_tree_depth": 0,
    "max_tree_depth": 0,
    "min_tree_depth": 0,
    "ave_tree_nodes": 0,
    "max_tree_nodes": 0,
    "min_tree_nodes": 0,
    "ave_fitness": 0,
    "best_fitness": 0,
    "time_taken": 0,
    "total_time": 0,
    "time_adjust": 0,
}


def get_stats(individuals, end=False):
    """
    Generate the statistics for an evolutionary run. Save statistics to
    utilities.trackers.stats_list. Print statistics. Save fitness plot
    information.

    :param individuals: A population of individuals for which to generate
    statistics.
    :param end: Boolean flag for indicating the end of an evolutionary run.
    :return: Nothing.
    """

    if hasattr(params["FITNESS_FUNCTION"], "multi_objective"):
        # Multiple objective optimisation is being used.

        # Remove fitness stats from the stats dictionary.
        stats.pop("best_fitness", None)
        stats.pop("ave_fitness", None)

        # Update stats.
        get_moo_stats(individuals, end)

    else:
        # Single objective optimisation is being used.
        get_soo_stats(individuals, end)

    if (
        params["SAVE_STATE"]
        and not params["DEBUG"]
        and stats["gen"] % params["SAVE_STATE_STEP"] == 0
    ):
        # Save the state of the current evolutionary run.
        create_state(individuals)


def get_soo_stats(individuals, end):
    """
    Generate the statistics for an evolutionary run with a single objective.
    Save statistics to utilities.trackers.stats_list. Print statistics. Save
    fitness plot information.

    :param individuals: A population of individuals for which to generate
    statistics.
    :param end: Boolean flag for indicating the end of an evolutionary run.
    :return: Nothing.
    """

    # Get best individual.
    best = max(individuals)

    if not trackers.best_ever or best > trackers.best_ever:
        # Save best individual in trackers.best_ever.
        trackers.best_ever = best

    if end or params["VERBOSE"] or not params["DEBUG"]:
        # Update all stats.
        update_stats(individuals, end)

    # Save fitness plot information
    if params["SAVE_PLOTS"] and not params["DEBUG"]:
        if not end:
            trackers.best_fitness_list.append(trackers.best_ever.fitness)

        if params["VERBOSE"] or end:
            save_plot_from_data(trackers.best_fitness_list, "best_fitness")

    # Print statistics
    if params["VERBOSE"] and not end:
        print_generation_stats()

    elif not params["SILENT"]:
        # Print simple display output.
        perc = stats["gen"] / (params["GENERATIONS"] + 1) * 100
        stdout.write("Evolution: %d%% complete\r" % perc)
        stdout.flush()

    # Generate test fitness on regression problems
    if hasattr(params["FITNESS_FUNCTION"], "training_test") and end:

        # Save training fitness.
        trackers.best_ever.training_fitness = copy(trackers.best_ever.fitness)

        # Evaluate test fitness.
        trackers.best_ever.test_fitness = params["FITNESS_FUNCTION"](
            trackers.best_ever, dist="test"
        )

        # Set main fitness as training fitness.
        trackers.best_ever.fitness = trackers.best_ever.training_fitness

    # Save stats to list.
    if params["VERBOSE"] or (not params["DEBUG"] and not end):
        trackers.stats_list.append(copy(stats))

    # Save stats to file.
    if not params["DEBUG"]:

        if stats["gen"] == 0:
            save_stats_headers(stats)

        save_stats_to_file(stats, end)

        if params["SAVE_ALL"]:
            save_best_ind_to_file(stats, trackers.best_ever, end, stats["gen"])

        elif params["VERBOSE"] or end:
            save_best_ind_to_file(stats, trackers.best_ever, end)

    if end and not params["SILENT"]:
        print_final_stats()


def get_moo_stats(individuals, end):
    """
    Generate the statistics for an evolutionary run with multiple objectives.
    Save statistics to utilities.trackers.stats_list. Print statistics. Save
    fitness plot information.

    :param individuals: A population of individuals for which to generate
    statistics.
    :param end: Boolean flag for indicating the end of an evolutionary run.
    :return: Nothing.
    """

    # Compute the pareto front metrics for the population.
    pareto = compute_pareto_metrics(individuals)

    # Save first front in trackers. Sort arbitrarily along first objective.
    trackers.best_ever = sorted(pareto.fronts[0], key=lambda x: x.fitness[0])

    # Store stats about pareto fronts.
    stats["pareto_fronts"] = len(pareto.fronts)
    stats["first_front"] = len(pareto.fronts[0])

    if end or params["VERBOSE"] or not params["DEBUG"]:
        # Update all stats.
        update_stats(individuals, end)

    # Save fitness plot information
    if params["SAVE_PLOTS"] and not params["DEBUG"]:

        # Initialise empty array for fitnesses for all inds on first pareto
        # front.
        all_arr = [[] for _ in range(params["FITNESS_FUNCTION"].num_obj)]

        # Generate array of fitness values.
        fitness_array = [ind.fitness for ind in trackers.best_ever]

        # Add paired fitnesses to array for graphing.
        for fit in fitness_array:
            for o in range(params["FITNESS_FUNCTION"].num_obj):
                all_arr[o].append(fit[o])

        if not end:
            trackers.first_pareto_list.append(all_arr)

            # Append empty array to best fitness list.
            trackers.best_fitness_list.append([])

            # Get best fitness for each objective.
            for o, ff in enumerate(params["FITNESS_FUNCTION"].fitness_functions):

                # Get sorted list of all fitness values for objective "o"
                fits = sorted(all_arr[o], reverse=ff.maximise)

                # Append best fitness to trackers list.
                trackers.best_fitness_list[-1].append(fits[0])

        if params["VERBOSE"] or end:

            # Plot best fitness for each objective.
            for o, ff in enumerate(params["FITNESS_FUNCTION"].fitness_functions):
                to_plot = [i[o] for i in trackers.best_fitness_list]

                # Plot fitness data for objective o.
                plotname = ff.__class__.__name__ + str(o)

                save_plot_from_data(to_plot, plotname)

            # TODO: PonyGE2 can currently only plot moo problems with 2 objectives.
            # Check that the number of fitness objectives is not greater than 2
            if params["FITNESS_FUNCTION"].num_obj > 2:
                s = (
                    "stats.stats.get_moo_stats\n"
                    "Warning: Plotting of more than 2 simultaneous "
                    "objectives is not yet enabled in PonyGE2."
                )
                print(s)

            else:
                save_pareto_fitness_plot()

    # Print statistics
    if params["VERBOSE"] and not end:
        print_generation_stats()
        print_first_front_stats()

    elif not params["SILENT"]:
        # Print simple display output.
        perc = stats["gen"] / (params["GENERATIONS"] + 1) * 100
        stdout.write("Evolution: %d%% complete\r" % perc)
        stdout.flush()

    # Generate test fitness on regression problems
    if hasattr(params["FITNESS_FUNCTION"], "training_test") and end:

        for ind in trackers.best_ever:
            # Iterate over all individuals in the first front.

            # Save training fitness.
            ind.training_fitness = copy(ind.fitness)

            # Evaluate test fitness.
            ind.test_fitness = params["FITNESS_FUNCTION"](ind, dist="test")

            # Set main fitness as training fitness.
            ind.fitness = ind.training_fitness

    # Save stats to list.
    if params["VERBOSE"] or (not params["DEBUG"] and not end):
        trackers.stats_list.append(copy(stats))

    # Save stats to file.
    if not params["DEBUG"]:

        if stats["gen"] == 0:
            save_stats_headers(stats)

        save_stats_to_file(stats, end)

        if params["SAVE_ALL"]:
            save_first_front_to_file(stats, end, stats["gen"])

        elif params["VERBOSE"] or end:
            save_first_front_to_file(stats, end)

    if end and not params["SILENT"]:
        print_final_moo_stats()


def update_stats(individuals, end):
    """
    Update all stats in the stats dictionary.

    :param individuals: A population of individuals.
    :param end: Boolean flag for indicating the end of an evolutionary run.
    :return: Nothing.
    """

    if not end:
        # Time Stats
        trackers.time_list.append(time() - stats["time_adjust"])
        stats["time_taken"] = trackers.time_list[-1] - trackers.time_list[-2]
        stats["total_time"] = trackers.time_list[-1] - trackers.time_list[0]

    # Population Stats
    stats["total_inds"] = params["POPULATION_SIZE"] * (stats["gen"] + 1)
    stats["runtime_error"] = len(trackers.runtime_error_cache)
    if params["CACHE"]:
        stats["unique_inds"] = len(trackers.cache)
        stats["unused_search"] = 100 - \
            stats["unique_inds"] / stats["total_inds"] * 100

    # Genome Stats
    genome_lengths = [len(i.genome) for i in individuals]
    stats["max_genome_length"] = np.nanmax(genome_lengths)
    stats["ave_genome_length"] = np.nanmean(genome_lengths)
    stats["min_genome_length"] = np.nanmin(genome_lengths)

    # Used Codon Stats
    codons = [i.used_codons for i in individuals]
    stats["max_used_codons"] = np.nanmax(codons)
    stats["ave_used_codons"] = np.nanmean(codons)
    stats["min_used_codons"] = np.nanmin(codons)

    # Tree Depth Stats
    depths = [i.depth for i in individuals]
    stats["max_tree_depth"] = np.nanmax(depths)
    stats["ave_tree_depth"] = np.nanmean(depths)
    stats["min_tree_depth"] = np.nanmin(depths)

    # Tree Node Stats
    nodes = [i.nodes for i in individuals]
    stats["max_tree_nodes"] = np.nanmax(nodes)
    stats["ave_tree_nodes"] = np.nanmean(nodes)
    stats["min_tree_nodes"] = np.nanmin(nodes)

    if not hasattr(params["FITNESS_FUNCTION"], "multi_objective"):
        # Fitness Stats
        fitnesses = [i.fitness for i in individuals]
        stats["ave_fitness"] = np.nanmean(fitnesses, axis=0)
        stats["best_fitness"] = trackers.best_ever.fitness


def print_generation_stats():
    """
    Print the statistics for the generation and individuals.

    :return: Nothing.
    """

    print("______\n")
    for stat in sorted(stats.keys()):
        print(" ", stat, ": \t", stats[stat])
    print("\n")


def print_first_front_stats():
    """
    Stats printing for the first pareto front for multi-objective optimisation.

    :return: Nothing.
    """

    print("  first front fitnesses :")
    for ind in trackers.best_ever:
        print("\t  ", ind.fitness)


def print_final_stats():
    """
    Prints a final review of the overall evolutionary process.

    :return: Nothing.
    """

    if hasattr(params["FITNESS_FUNCTION"], "training_test"):
        print("\n\nBest:\n  Training fitness:\t",
              trackers.best_ever.training_fitness)
        print("  Test fitness:\t\t", trackers.best_ever.test_fitness)
    else:
        print("\n\nBest:\n  Fitness:\t", trackers.best_ever.fitness)

    print("  Phenotype:", trackers.best_ever.phenotype)
    print("  Genome:", trackers.best_ever.genome)
    print_generation_stats()


def print_final_moo_stats():
    """
    Prints a final review of the overall evolutionary process for
    multi-objective problems.

    :return: Nothing.
    """

    print("\n\nFirst Front:")
    for ind in trackers.best_ever:
        print(" ", ind)
    print_generation_stats()



#!/usr/bin/env python
"""
@author: Jonathan Byrne
@17/01/18 11:09
"""

import numpy as np

from autooc.fitness.base_ff_classes.base_ff import base_ff


class singlefit_multiobj(base_ff):
    """
    An example of a single fitness class that generates
    two fitness values for multi-objective optimisation
    """

    maximise = True
    multi_objective = True

    def __init__(self):

        # Initialise base fitness function class.
        super().__init__()

        # Set list of individual fitness functions.
        self.num_obj = 2
        dummyfit = base_ff()
        dummyfit.maximise = True
        self.fitness_functions = [dummyfit, dummyfit]
        self.default_fitness = [float("nan"), float("nan")]

    def evaluate(self, ind, **kwargs):
        """Dummy fitness function that generates 2 fitness values"""
        phenotype = ind.phenotype
        fitness = 0
        settings = {}

        # try:
        #     exec(phenotype, settings)
        # except Exception as e:
        #     fitness = self.default_fitness
        #     print("Error", e)

        # Using dummy fitness values for the moment.
        x = np.random.pareto(4, 2)
        fitness = [x[0], x[1]]

        return fitness

    @staticmethod
    def value(fitness_vector, objective_index):
        """
        This is a static method required by NSGA-II for sorting populations
        based on a given fitness function, or for returning a given index of a
        population based on a given fitness function.

        :param fitness_vector: A vector/list of fitnesses.
        :param objective_index: The index of the desired fitness.
        :return: The fitness at the objective index of the fitness vector.
        """

        if not isinstance(fitness_vector, list):
            return float("inf")

        return fitness_vector[objective_index]



from multiprocessing import cpu_count
from os import getcwd, path
from socket import gethostname

hostname = gethostname().split(".")
machine_name = hostname[0]


"""Algorithm parameters"""
params = {
    # Set default step and search loop functions
    "SEARCH_LOOP": "search_loop",
    "STEP": "step",
    # Evolutionary Parameters
    "POPULATION_SIZE": 500,
    "GENERATIONS": 50,
    "HILL_CLIMBING_HISTORY": 1000,
    "SCHC_COUNT_METHOD": "count_all",
    # Set optional experiment name
    "EXPERIMENT_NAME": "AutoOC",
    # Set default number of runs to be done.
    # ONLY USED WITH EXPERIMENT MANAGER.
    "RUNS": 1,
    # Class of problem
    "FITNESS_FUNCTION": "supervised_learning.regression",
    # Select problem dataset
    "DATASET_TRAIN": None,
    "DATASET_TEST": None,
    "DATASET_DELIMITER": None,
    # Set grammar file
    # "GRAMMAR_FILE": "autoencoders.pybnf",
    # Set the number of depths permutations are calculated for
    # (starting from the minimum path of the grammar).
    # Mainly for use with the grammar analyser script.
    "PERMUTATION_RAMPS": 5,
    # Select error metric
    # "ERROR_METRIC": "reconstruction_error",
    # Optimise constants in the supervised_learning fitness function.
    "OPTIMIZE_CONSTANTS": False,
    # Specify target for target problems
    # "TARGET": "ponyge_rocks",
    # Set max sizes of individuals
    "MAX_TREE_DEPTH": 20,  # SET TO 90 DUE TO PYTHON EVAL() STACK LIMIT.
    # INCREASE AT YOUR OWN RISK.
    "MAX_TREE_NODES": None,
    "CODON_SIZE": 100000,
    "MAX_GENOME_LENGTH": 1000,
    "MAX_WRAPS": 0,
    # INITIALISATION
    # Set initialisation operator.
    "INITIALISATION": "PI_grow",
    # Set the maximum geneome length for initialisation.
    "INIT_GENOME_LENGTH": 200,
    # Set the maximum tree depth for initialisation.
    "MAX_INIT_TREE_DEPTH": 10,
    # Set the minimum tree depth for initialisation.
    "MIN_INIT_TREE_DEPTH": None,
    # SELECTION
    # Set selection operator.
    "SELECTION": "autooc.operators.selection.tournament",
    # For tournament selection
    "TOURNAMENT_SIZE": 2,
    # For truncation selection
    "SELECTION_PROPORTION": 0.5,
    # Allow for selection of invalid individuals during selection process.
    "INVALID_SELECTION": False,
    # OPERATOR OPTIONS
    # Boolean flag for selecting whether or not mutation is confined to
    # within the used portion of the genome. Default set to True.
    "WITHIN_USED": True,
    # CROSSOVER
    # Set crossover operator.
    "CROSSOVER": "autooc.operators.crossover.variable_onepoint",
    # Set crossover probability.
    "CROSSOVER_PROBABILITY": 0.75,
    # Prevents crossover from generating invalids.
    "NO_CROSSOVER_INVALIDS": False,
    # MUTATION
    # Set mutation operator.
    "MUTATION": "autooc.operators.mutation.int_flip_per_codon",
    # Set mutation probability (None defaults to 1 over the length of
    # the genome for each codon)
    "MUTATION_PROBABILITY": None,
    # Set number of mutation events
    "MUTATION_EVENTS": 1,
    # Prevents mutation from generating invalids.
    "NO_MUTATION_INVALIDS": False,
    # REPLACEMENT
    # Set replacement operator.
    "REPLACEMENT": "autooc.operators.replacement.generational",
    # Set elite size.
    "ELITE_SIZE": None,
    # DEBUGGING
    # Use this to turn on debugging mode. This mode doesn't write any files
    # and should be used when you want to test new methods.
    "DEBUG": False,
    # PRINTING
    # Use this to print out basic statistics for each generation to the
    # command line.
    "VERBOSE": True,
    # Use this to prevent anything being printed to the command line.
    "SILENT": False,
    # SAVING
    # Save the phenotype of the best individual from each generation. Can
    # generate a lot of files. DEBUG must be False.
    "SAVE_ALL": False,
    # Save a plot of the evolution of the best fitness result for each
    # generation.
    "SAVE_PLOTS": True,
    # MULTIPROCESSING
    # Multi-core parallel processing of phenotype evaluations.
    "MULTICORE": False,
    # Set the number of cpus to be used for multiprocessing
    "CORES": cpu_count(),
    # STATE SAVING/LOADING
    # Save the state of the evolutionary run every generation. You can
    # specify how often you want to save the state with SAVE_STATE_STEP.
    "SAVE_STATE": False,
    # Specify how often the state of the current evolutionary run is
    # saved (i.e. every n-th generation). Requires int value.
    "SAVE_STATE_STEP": 1,
    # Load an evolutionary run from a saved state. You must specify the
    # full file path to the desired state file. Note that state files have
    # no file type.
    "LOAD_STATE": None,
    # SEEDING
    # Specify a list of PonyGE2 individuals with which to seed the initial
    # population.
    "SEED_INDIVIDUALS": [],
    # Specify a target seed folder in the 'seeds' directory that contains a
    # population of individuals with which to seed a run.
    "TARGET_SEED_FOLDER": "",
    # Set a target phenotype string for reverse mapping into a GE
    # individual
    "REVERSE_MAPPING_TARGET": None,
    # Set Random Seed for all Random Number Generators to be used by
    # PonyGE2, including the standard Python RNG and the NumPy RNG.
    "RANDOM_SEED": None,
    # CACHING
    # The cache tracks unique individuals across evolution by saving a
    # string of each phenotype in a big list of all phenotypes. Saves all
    # fitness information on each individual. Gives you an idea of how much
    # repetition is in standard GE/GP.
    "CACHE": True,
    # Uses the cache to look up the fitness of duplicate individuals. CACHE
    # must be set to True if you want to use this.
    "LOOKUP_FITNESS": False,
    # Uses the cache to give a bad fitness to duplicate individuals. CACHE
    # must be True if you want to use this (obviously)
    "LOOKUP_BAD_FITNESS": False,
    # Removes duplicate individuals from the population by replacing them
    # with mutated versions of the original individual. Hopefully this will
    # encourage diversity in the population.
    "MUTATE_DUPLICATES": False,
    # MULTI-AGENT Parameters
    # True or False for multi-agent
    "MULTIAGENT": False,
    # Agent Size. Number of agents having their own copy of genetic material
    "AGENT_SIZE": 100,
    # Interaction Probability. How frequently the agents can interaction with each other
    "INTERACTION_PROBABILITY": 0.5,
    # OTHER
    # Set machine name (useful for doing multiple runs)
    "MACHINE": machine_name,
    # EPOCHS for model training
    "EPOCHS": 50,
    "TRAINING_TIME": 0,
    # "NORMAL_TRAIN_DATA": None,
    # "NORMAL_TEST_DATA": None,
    # "TEST_DATA": None,
    # "INPUT_SHAPE": None,
}


def load_params(file_name):
    """
    Load in a params text file and set the params dictionary directly.

    :param file_name: The name/location of a parameters file.
    :return: Nothing.
    """
    print(file_name)
    try:
        open(file_name, "r")
    except FileNotFoundError:
        s = (
            "algorithm.parameters.load_params\n"
            "Error: Parameters file not found.\n"
            "       Ensure file extension is specified, e.g. 'regression.txt'."
        )
        raise Exception(s)

    with open(file_name, "r") as parameters:
        # Read the whole parameters file.
        content = parameters.readlines()

        for line in [l for l in content if not l.startswith("#")]:

            # Parameters files are parsed by finding the first instance of a
            # colon.
            split = line.find(":")

            # Everything to the left of the colon is the parameter key,
            # everything to the right is the parameter value.
            key, value = line[:split], line[split + 1:].strip()

            # Evaluate parameters.
            try:
                value = eval(value)

            except:
                # We can't evaluate, leave value as a string.
                pass

            # Set parameter
            params[key] = value


def set_params(command_line_args, create_files=True):
    """
    This function parses all command line arguments specified by the user.
    If certain parameters are not set then defaults are used (e.g. random
    seeds, elite size). Sets the correct imports given command line
    arguments. Sets correct grammar file and fitness function. Also
    initialises save folders and tracker lists in utilities.trackers.

    :param command_line_args: Command line arguments specified by the user.
    :return: Nothing.
    """
    from autooc.representation import grammar
    from autooc.utilities.algorithm.command_line_parser import parse_cmd_args
    from autooc.utilities.algorithm.initialise_run import (
        initialise_run_params, set_param_imports)
    from autooc.utilities.fitness.math_functions import return_one_percent
    from autooc.utilities.stats import clean_stats, trackers

    cmd_args, unknown = parse_cmd_args(command_line_args)

    if unknown:
        # We currently do not parse unknown parameters. Raise error.
        s = (
            "algorithm.parameters.set_params\nError: "
            "unknown parameters: %s\nYou may wish to check the spelling, "
            "add code to recognise this parameter, or use "
            "--extra_parameters" % str(unknown)
        )
        raise Exception(s)

    # LOAD PARAMETERS FILE
    # NOTE that the parameters file overwrites all previously set parameters.
    if "PARAMETERS" in cmd_args:
        load_params(path.join(getcwd(), "autooc",
                    "parameters", cmd_args["PARAMETERS"]))

    # Join original params dictionary with command line specified arguments.
    # NOTE that command line arguments overwrite all previously set parameters.
    params.update(cmd_args)

    if params["LOAD_STATE"]:
        # Load run from state.
        from autooc.utilities.algorithm.state import load_state

        # Load in state information.
        individuals = load_state(params["LOAD_STATE"])

        # Set correct search loop.
        from autooc.algorithm.search_loop import search_loop_from_state

        params["SEARCH_LOOP"] = search_loop_from_state

        # Set population.
        setattr(trackers, "state_individuals", individuals)

    else:
        if params["REPLACEMENT"].split(".")[-1] == "steady_state":
            # Set steady state step and replacement.
            params["STEP"] = "steady_state_step"
            params["GENERATION_SIZE"] = 2

        else:
            # Elite size is set to either 1 or 1% of the population size,
            # whichever is bigger if no elite size is previously set.
            if params["ELITE_SIZE"] is None:
                params["ELITE_SIZE"] = return_one_percent(
                    1, params["POPULATION_SIZE"])

            # Set the size of a generation
            params["GENERATION_SIZE"] = params["POPULATION_SIZE"] - \
                params["ELITE_SIZE"]

        # Initialise run lists and folders before we set imports.r
        # initialise_run_params(create_files)

        # Set correct param imports for specified function options, including
        # error metrics and fitness functions.
        set_param_imports()

        # Clean the stats dict to remove unused stats.
        clean_stats.clean_stats()

        # Set GENOME_OPERATIONS automatically for faster linear operations.
        if (
            params["CROSSOVER"].representation == "subtree"
            or params["MUTATION"].representation == "subtree"
        ):
            params["GENOME_OPERATIONS"] = False
        else:
            params["GENOME_OPERATIONS"] = True

        # Ensure correct operators are used if multiple fitness functions used.
        if hasattr(params["FITNESS_FUNCTION"], "multi_objective"):

            # Check that multi-objective compatible selection is specified.
            if not hasattr(params["SELECTION"], "multi_objective"):
                s = (
                    "algorithm.parameters.set_params\n"
                    "Error: multi-objective compatible selection "
                    "operator not specified for use with multiple "
                    "fitness functions."
                )
                raise Exception(s)

            if not hasattr(params["REPLACEMENT"], "multi_objective"):

                # Check that multi-objective compatible replacement is
                # specified.
                if not hasattr(params["REPLACEMENT"], "multi_objective"):
                    s = (
                        "algorithm.parameters.set_params\n"
                        "Error: multi-objective compatible replacement "
                        "operator not specified for use with multiple "
                        "fitness functions."
                    )
                    raise Exception(s)

        # Parse grammar file and set grammar class.
        params["BNF_GRAMMAR"] = grammar.Grammar(
            params["GRAMMAR_FILE"]
            #path.join("autooc", "grammars", params["GRAMMAR_FILE"])
            #path.join(path.dirname(path.abspath(__file__)), "..", "grammars", params["GRAMMAR_FILE"])
        )

        # Population loading for seeding runs (if specified)
        if params["TARGET_SEED_FOLDER"]:

            # Import population loading function.
            from autooc.operators.initialisation import load_population

            # A target folder containing seed individuals has been given.
            params["SEED_INDIVIDUALS"] = load_population(
                params["TARGET_SEED_FOLDER"])

        elif params["REVERSE_MAPPING_TARGET"]:
            # A single seed phenotype has been given. Parse and run.

            # Import GE LR Parser.
            from autooc.scripts import GE_LR_parser

            # Parse seed individual and store in params.
            params["SEED_INDIVIDUALS"] = [GE_LR_parser.main()]



def python_filter(txt):
    """Create correct python syntax.

    We use {: and :} as special open and close brackets, because
    it's not possible to specify indentation correctly in a BNF
    grammar without this type of scheme."""

    indent_level = 0
    tmp = txt[:]
    i = 0
    while i < len(tmp):
        tok = tmp[i: i + 2]
        if tok == "{:":
            indent_level += 1
        elif tok == ":}":
            indent_level -= 1
        tabstr = "\n" + "  " * indent_level
        if tok == "{:" or tok == ":}":
            tmp = tmp.replace(tok, tabstr, 1)
        i += 1
    # Strip superfluous blank lines.
    txt = "\n".join([line for line in tmp.split("\n") if line.strip() != ""])
    return txt



import re
from copy import copy
from itertools import zip_longest

from autooc.algorithm.parameters import params
from autooc.representation import individual, tree
from autooc.utilities.representation.check_methods import (generate_codon,
                                                          get_output)
from autooc.utilities.stats import trackers


def combine_snippets():
    """
    As the snippets repository grows, we can start to combine
    neighboring snippets to build bigger snippets. Eventually we hope this
    can just build the perfect solution. Iteratively builds snippets until
    no more snippets can be built form the current library.

    :return: Nothing.
    """

    # Find the number of snippets at T.
    original_snippets = sorted(trackers.snippets.keys())

    # Perform first pass of reduction.
    reduce_trees()

    # Delete obsolete snippets.
    remove_old_snippets()

    # Get new snippets list.
    updated_snippets = sorted(trackers.snippets.keys())

    # Initialise counter for reduction iterations.
    no_passes = 1

    while updated_snippets != original_snippets:
        # Keep reducing snippets until no more reductions can be made.

        # Save old T+1
        pre_updated_snippets = copy(updated_snippets)

        # Perform reduction.
        reduce_trees()

        # Delete obsolete snippets.
        remove_old_snippets()

        # Get new snippets list.
        updated_snippets = sorted(trackers.snippets.keys())

        # Set new T as old T+1
        original_snippets = pre_updated_snippets

        # Increment counter
        no_passes += 1


def reduce_trees():
    """
    Iterates through all snippets in the snippets dictionary and reduces
    snippets to make larger snippets.

    :return: Nothing.
    """

    # Get list of all reduction NTs.
    reduce_NTs = params["BNF_GRAMMAR"].concat_NTs

    # Sort snippets keys.
    sorted_keys = sorted(
        [
            [get_num_from_str(snippet), get_NT_from_str(snippet), snippet]
            for snippet in trackers.snippets.keys()
        ]
    )

    # Iterate over all snippets.
    for snippet_info in sorted_keys:
        # print(snippet_info)

        # Get current snippet.
        snippet = snippet_info[2]

        # Find current snippet info.
        NT = snippet_info[1]

        # Get indexes of the current snippet
        indexes = snippet_info[0]
        start, end = indexes[0], indexes[1]

        # Find if the snippet root (NT) exists anywhere in the
        # reduction NTs.
        if NT in reduce_NTs:

            for reduce in reduce_NTs[NT]:
                # Now we're searching for a specific subset of keys in the
                # snippets dictionary.

                # Generate list of only the desired Non Terminals.
                NTs = reduce[2]

                if len(NTs) == 1:
                    # This choice leads directly to the parent, check if parent
                    # snippet already exists.

                    # Child is current snippet.
                    child = [[snippet, trackers.snippets[snippet]]]

                    generate_key_and_check(start, end, reduce, child)

                else:
                    # Find the index of the snippet root in the current
                    # reduction production choice.
                    NT_locs = [i for i, x in enumerate(NTs) if x[0] == NT]

                    for loc in NT_locs:
                        # We want to check each possible reduction option.
                        # if speck:

                        # Set where the original snippet starts and ends on
                        # the target string.
                        if loc == 0:
                            # The current snippet is at the start of the
                            # reduction attempt.
                            aft, pre = end, None

                        elif start == 0 and loc != 0:
                            # The current snippet is at the start of the target
                            # string, but we are trying to reduce_trees it with
                            # something before it.
                            break

                        elif end == len(params["TARGET"]) and loc != NT_locs[-1]:
                            # The current snippet is at the end of the target
                            # string, but we are trying to reduce_trees it with
                            # something after it.
                            break

                        elif loc == len(NTs):
                            # The current snippet is at the end of the
                            # reduction attempt.
                            aft, pre = None, start

                        else:
                            # The current snippet is in the middle of the
                            # reduction attempt.
                            aft, pre = end, start

                        alt_cs = list(range(len(NTs)))

                        # Initialise a list of children to be reduced.
                        children = [[] for _ in range(len(NTs))]

                        # Set original snippet into children.
                        children[loc] = [snippet, trackers.snippets[snippet]]

                        # Generate ordered list of alternating indexes of Ts
                        # and NTs to reduce_trees with a given original NT.
                        b = zip_longest(alt_cs[loc:], reversed(alt_cs[:loc]))
                        alt_cs = [x for x in list(sum(b, ())) if x is not None]
                        alt_cs.remove(loc)

                        def check_reductions(alt_cs, pre, aft, idx, children):
                            """
                            Given a list of the indexes of potential children, find
                            snippets which match adjacent portions of the target
                            string that can be used as these children.

                            :param alt_cs: An ordered list of indexes of
                            potential Terminals or Non Terminals to
                            reduce_trees.
                            :param pre: The start index of the overall snippet
                            on the target string.
                            :param aft: The end index of the overall snippet on
                            the target string.
                            :param children: A list of children to be reduced.
                            :param idx: The index of the current child.
                            :return: The same inputs, in the same order.
                            """

                            if idx < len(alt_cs):

                                # Take the next available unexpanded item from the
                                # list.
                                child_idx = alt_cs[idx]
                                child = NTs[child_idx]

                                # Increment counter.
                                idx += 1

                                if pre is None:
                                    # Then we are starting with an NT which
                                    # directly follows the current NT.

                                    if child[1] == "T":
                                        # Then we can immediately check if this
                                        # reduction T is legally allowed
                                        # follow the current NT in the target
                                        # string.

                                        # Get output to check for match.
                                        check = child[0]

                                        # Get portion of target string to match.
                                        end_point = aft + len(check)

                                        target = params["TARGET"][aft:end_point]

                                        if target == check:
                                            # The current terminal matches the same
                                            # location on the target string.

                                            # Generate fake key for snippets dict.
                                            key = str([aft, end_point])

                                            # Increment aft phenotype counter.
                                            aft += len(check)

                                            # Create new tree from this terminal.
                                            T_tree = tree.Tree(check, None)

                                            # Add to children.
                                            children[child_idx] = [key, T_tree]

                                            if alt_cs:
                                                # Recurse to find the next piece of
                                                # the puzzle.
                                                check_reductions(
                                                    alt_cs, pre, aft, idx, children
                                                )

                                    elif child[1] == "NT":
                                        # Check to see if there are any snippets
                                        # which can be reduced to the current
                                        # block.

                                        # Find all potential snippets which
                                        # match our criteria.
                                        matches = [
                                            v
                                            for v in sorted_keys
                                            if v[0][0] == aft and v[1] == child[0]
                                        ]

                                        # Create a copy of this marker otherwise
                                        # each match will over-write it.
                                        aft_c = copy(aft)

                                        for match in matches:
                                            # Iterate over all potential matches.

                                            # Calculate length of match string.
                                            str_len = match[0]

                                            # Increment appropriate phenotype
                                            # counter.
                                            aft_c = aft + \
                                                str_len[1] - str_len[0]

                                            # Add to children.
                                            children[child_idx] = [
                                                match[2],
                                                trackers.snippets[match[2]],
                                            ]

                                            if alt_cs:
                                                # Recurse to find the next piece of
                                                # the puzzle.
                                                check_reductions(
                                                    alt_cs, pre, aft_c, idx, children
                                                )

                                elif aft is None:
                                    # Then we are starting with an NT which
                                    # directly precedes the current NT.

                                    if child[1] == "T":
                                        # Then we can immediately check if this
                                        # reduction T is legally allowed
                                        # follow the current NT in the target
                                        # string.

                                        # Get output to check for match.
                                        check = child[0]

                                        # Get portion of target string to match.
                                        start_point = pre - len(check)

                                        target = params["TARGET"][start_point:pre]

                                        if target == check:
                                            # The current terminal matches the same
                                            # location on the target string.

                                            # Generate fake key for snippets dict.
                                            key = str([start_point, pre])

                                            # Increment pre phenotype counter.
                                            pre -= len(check)

                                            # Create new tree from this terminal.
                                            T_tree = tree.Tree(check, None)

                                            # Add to children.
                                            children[child_idx] = [key, T_tree]

                                            if alt_cs:
                                                # Recurse to find the next piece of
                                                # the puzzle.
                                                check_reductions(
                                                    alt_cs, pre, aft, idx, children
                                                )

                                    elif child[1] == "NT":
                                        # Check to see if there are any snippets
                                        # which can be reduced to the current
                                        # block.

                                        # Find all potential snippets which
                                        # match our criteria.
                                        matches = [
                                            v
                                            for v in sorted_keys
                                            if v[0][1] == pre and v[1] == child[0]
                                        ]

                                        # Create a copy of this marker otherwise
                                        # each match will over-write it.
                                        pre_c = copy(pre)

                                        for match in matches:
                                            # Iterate over all potential matches.

                                            # Calculate length of match string.
                                            str_len = match[0]

                                            # Increment appropriate phenotype
                                            # counter.
                                            pre_c = pre - \
                                                str_len[1] + str_len[0]

                                            # Add to children.
                                            children[child_idx] = [
                                                match[2],
                                                trackers.snippets[match[2]],
                                            ]

                                            if alt_cs:
                                                # Recurse to find the next piece of
                                                # the puzzle.
                                                check_reductions(
                                                    alt_cs, pre_c, aft, idx, children
                                                )

                                else:
                                    # Our starting NT is somewhere in the middle
                                    # of the proposed reduction.

                                    if child[1] == "T":
                                        # Then we can immediately check if this
                                        # reduction T is legally allowed be
                                        # where it wants to be.

                                        # Get output to check for match.
                                        check = child[0]

                                        # Get portion of target string to match.
                                        if child_idx > loc:
                                            # This T comes after the original NT.
                                            start_point = aft
                                            end_point = start_point + \
                                                len(check)

                                        else:
                                            # This T comes before the original NT.
                                            start_point = pre - len(check)
                                            end_point = pre

                                        target = params["TARGET"][start_point:end_point]

                                        if target == check:
                                            # The current terminal matches the same
                                            # location on the target string.

                                            # Increment appropriate phenotype
                                            # counter.
                                            if child_idx > loc:
                                                # This T comes after the original
                                                # NT.
                                                aft += len(check)

                                            else:
                                                # This T comes before the original
                                                # NT.
                                                pre -= len(check)

                                            # Generate fake key for snippets dict.
                                            key = str([start_point, end_point])

                                            # Create new tree from this terminal.
                                            T_tree = tree.Tree(check, None)

                                            # Add to children.
                                            children[child_idx] = [key, T_tree]

                                            if alt_cs:
                                                # Recurse to find the next piece of
                                                # the puzzle.
                                                check_reductions(
                                                    alt_cs, pre, aft, idx, children
                                                )

                                    elif child[1] == "NT":
                                        # Check to see if there are any snippets
                                        # which can be reduced to the current
                                        # block.

                                        # Get portion of target string to match.
                                        if child_idx > loc:
                                            # This NT comes after the original NT.
                                            matches = [
                                                v
                                                for v in sorted_keys
                                                if v[0][0] == aft and v[1] == child[0]
                                            ]

                                        else:
                                            # This NT comes before the original NT.
                                            matches = [
                                                v
                                                for v in sorted_keys
                                                if v[0][1] == pre and v[1] == child[0]
                                            ]

                                        # Create copies of this markers otherwise
                                        # each match will over-write them.
                                        pre_c, aft_c = copy(pre), copy(aft)

                                        for match in matches:
                                            # Iterate over all potential matches.

                                            # Calculate length of match string.
                                            str_len = match[0]

                                            # Increment appropriate phenotype
                                            # counter.
                                            if child_idx > loc:
                                                # This NT comes after the original
                                                # NT.
                                                aft_c = aft + \
                                                    str_len[1] - str_len[0]

                                            else:
                                                # This NT comes before the original
                                                # NT.
                                                pre_c = pre - \
                                                    str_len[1] + str_len[0]

                                            # Add to children.
                                            children[child_idx] = [
                                                match[2],
                                                trackers.snippets[match[2]],
                                            ]

                                            if alt_cs:
                                                # Recurse to find the next piece of
                                                # the puzzle.
                                                check_reductions(
                                                    alt_cs, pre_c, aft_c, idx, children
                                                )

                            elif all([i != [] for i in children]):
                                # We have compiled a full set of potential
                                # children to reduce_trees. Generate a key and check
                                # if it exists.
                                generate_key_and_check(
                                    pre, aft, reduce, children)

                        # Check whether a reduction can be performed.
                        check_reductions(alt_cs, pre, aft, 0, children)


def generate_key_and_check(pre, aft, reduce, children):
    """
    Will generate a snippet key and check if it exists in the repository. If
    snippet is not in the repository, adds it.

    :param pre: The start index of the overall snippet on the target string.
    :param aft: The end index of the overall snippet on the target string.
    :param reduce: The information necessary to reduce_trees a list of snippets.
    Includes the root NT that will form the new root node of the reduced
    snippet.
    :param children: A dictionary containing the derivation trees of all
    components of the snippet to be reduced.
    :return: Nothing.
    """

    if pre is None:
        pre = get_num_from_str(children[0][0])[0]

    if aft is None:
        aft = get_num_from_str(children[-1][0])[1]

    # Generate key for proposed reduction
    new_key = " ".join([str([pre, aft]), reduce[1]])

    if new_key in trackers.snippets or new_key in trackers.deleted_snippets:
        # No need to reduce_trees as a perfectly good
        # solution already exists.
        pass

    else:
        # Create list of children.
        children = [i[1].__copy__() for i in children]

        # We can generate a new snippet by reducing
        # two existing snippets.
        create_snippet(reduce[1], children, reduce[0], new_key)

    return new_key, pre, aft


def remove_old_snippets():
    """
    Iterate over the snippets repository and remove snippets which are
    sub-trees of larger snippets as these snippets are useless now.

    :return: Nothing.
    """

    for snippet in sorted(trackers.snippets.keys()):
        # Iterate over all snippets.

        if snippet in trackers.snippets:
            delete_snippet(trackers.snippets[snippet])


def delete_snippet(self):
    """
    Given a tree structure, dive down through the tree recursively and delete
    all child snippets.

    :param self: A parse tree.
    :return: Nothing.
    """

    if (
        self.parent
        and self.snippet
        and self.snippet in trackers.snippets
        and len(params["BNF_GRAMMAR"].concat_NTs[self.root]) == 1
    ):

        # Delete this snippet as it's (hopefully) useless now.
        del trackers.snippets[self.snippet]

        # Append deleted snippet to deleted snippets list.
        trackers.deleted_snippets.append(self.snippet)

    if self.children:
        # Recurse through all children.

        for child in self.children:
            delete_snippet(child)


def create_snippet(parent, children, choice, key):
    """
    Given a parent NT and a list of child trees, create a new tree that acts as
    the parent of the given children. Generates this tree as a snippet and
    adds the new snippet to the trackers.snippets library.

    :param parent: A non-terminal root.
    :param children: A list of derivation tree instances.
    :param choice: The chosen production choice.
    :param key: A new key for the trackers.snippets dictionary.
    :return: Nothing.
    """

    # Initialise new instance of the tree class to act as new snippet.
    new_tree = tree.Tree(parent, None)

    # Generate a codon to match the given production choice.
    new_tree.codon = generate_codon(parent, choice)

    # Save the snippet key of this tree.
    new_tree.snippet = key

    # Add the children to the new node
    for child in children:
        new_tree.children.append(child)

        # Set the parent of the child to the new node
        child.parent = new_tree

    # Add new snippet to snippets dictionary
    trackers.snippets[key] = new_tree


def get_num_from_str(string):
    """
    Given a string of a snippet, return the indexes of that snippet.

     in: '[1, 2] <RE>'

     out: [1, 2]

    :param string: A string defining a snippet.
    :return: The indexes of that snippet.
    """

    # Get index portion of string
    index = re.findall("\[\d+, \d+\]", string)

    return eval(index[0])


def get_NT_from_str(string):
    """
    Given a string of a snippet, return the NT of that snippet.

     in: '[1, 2] <RE>'

     out: '<RE>'

    :param string: A string defining a snippet.
    :return: The NT of that snippet.
    """

    # Get index portion of string
    index = re.findall("\<.+\>", string)
    return index[0]


def check_snippets_for_solution():
    """
    Check the snippets repository to see if we have built up the correct
    solution yet.

    :return: An individual representing the correct solution if it exists,
    otherwise None.
    """

    # Initialise None biggest snippet
    biggest_snippet = [0, None]

    for snippet in sorted(trackers.snippets.keys()):
        # Check each snippet to find the largest one.

        # Find length of snippet
        index = get_num_from_str(snippet)
        length = index[1] - index[0]

        if length > biggest_snippet[0]:
            # We have a new biggest snippet.
            biggest_snippet = [length, snippet]

    # Get the phenotype of the largest snippet
    largest_snippet = get_output(trackers.snippets[biggest_snippet[1]])

    if largest_snippet != params["REVERSE_MAPPING_TARGET"]:
        # The solution doesn't match the target string.

        # Get the location of the phenotype of the largest snippet on the
        # target string.
        largest_indexes = get_num_from_str(biggest_snippet[1])

        # Generate whitespace to position the phenotype accordingly.
        spaces = "".join([" " for _ in range(largest_indexes[0] - 1)])

        s = (
            "operators.subtree_parse.check_snippets_for_solution\n"
            "Error: Solution doesn't match the target string.\n"
            "       Target:   \t %s\n"
            "       Solution: \t %s %s\n"
            "       Check grammar file `%s` to ensure the grammar is capable"
            " of producing the exact target string."
            % (
                params["REVERSE_MAPPING_TARGET"],
                spaces,
                largest_snippet,
                params["GRAMMAR_FILE"],
            )
        )
        raise Exception(s)

    if largest_snippet == params["REVERSE_MAPPING_TARGET"]:
        # We have a perfect match

        # Generate individual that represents the perfect solution.
        ind = individual.Individual(
            None, trackers.snippets[biggest_snippet[1]])

        # Return ind.
        return ind



from os import path

from tensorflow.keras import Sequential
from tensorflow.keras.layers import Dense


def reverse(data_list):
    length = len(data_list)
    s = length
    new_list = [None] * length
    for item in data_list:
        s = s - 1
        new_list[s] = item
    return new_list


def get_model_from_encoder(encoder):
    # get model configs
    configs = encoder.get_config()
    # extract layers
    layers = configs.get("layers")
    # remove layers' names to prevent errors
    for layer in layers:
        if layer["config"]["name"] not in ["latent", "input"]:
            del layer["config"]["name"]
    # reverse layers (last one is the latent space, should not be repeated!)
    rev_layers = reverse(layers[1:-1])
    # concat layers
    all_layers = layers + rev_layers
    # create model
    ae = Sequential.from_config(all_layers)

    return ae


def get_model_from_encoder2(encoder):
    # get model configs
    configs = encoder.get_config()
    # extract layers
    layers = configs.get("layers")
    # get input layer shape
    current_shape = layers[0]["config"]["batch_input_shape"][1]
    # adjust layers' shapes
    for layer in layers:
        if "units" in layer["config"]:
            current_shape = int(
                current_shape * (layer["config"]["units"] / 100))
            if current_shape < 1:
                current_shape = 1
            layer["config"]["units"] = current_shape
            print(layer["config"]["units"] / 10)
    # remove layers' names to prevent errors
    for layer in layers:
        if layer["config"]["name"] not in ["latent", "input"]:
            del layer["config"]["name"]
    # reverse layers (last one is the latent space, should not be repeated!)
    rev_layers = reverse(layers[1:-1])
    # concat layers
    all_layers = layers + rev_layers
    # create model
    ae = Sequential.from_config(all_layers)
    return ae



from random import choice, randint, random

from autooc.algorithm.parameters import params
from autooc.representation import individual
from autooc.representation.derivation import generate_tree
from autooc.representation.latent_tree import (latent_tree_mutate,
                                              latent_tree_repair)
from autooc.utilities.representation.check_methods import check_ind


def mutation(pop):
    """
    Perform mutation on a population of individuals. Calls mutation operator as
    specified in params dictionary.

    :param pop: A population of individuals to be mutated.
    :return: A fully mutated population.
    """

    # Initialise empty pop for mutated individuals.
    new_pop = []

    # Iterate over entire population.
    for ind in pop:

        # If individual has no genome, default to subtree mutation.
        if not ind.genome and params["NO_MUTATION_INVALIDS"]:
            new_ind = subtree(ind)

        else:
            # Perform mutation.
            new_ind = params["MUTATION"](ind)

        # Check ind does not violate specified limits.
        check = check_ind(new_ind, "mutation")

        while check:
            # Perform mutation until the individual passes all tests.

            # If individual has no genome, default to subtree mutation.
            if not ind.genome and params["NO_MUTATION_INVALIDS"]:
                new_ind = subtree(ind)

            else:
                # Perform mutation.
                new_ind = params["MUTATION"](ind)

            # Check ind does not violate specified limits.
            check = check_ind(new_ind, "mutation")

        # Append mutated individual to population.
        new_pop.append(new_ind)

    return new_pop


def int_flip_per_codon(ind):
    """
    Mutate the genome of an individual by randomly choosing a new int with
    probability p_mut. Works per-codon. Mutation is performed over the
    effective length (i.e. within used codons, not tails) by default;
    within_used=False switches this off.

    :param ind: An individual to be mutated.
    :return: A mutated individual.
    """

    # Set effective genome length over which mutation will be performed.
    eff_length = get_effective_length(ind)

    if not eff_length:
        # Linear mutation cannot be performed on this individual.
        return ind

    # Set mutation probability. Default is 1 over the length of the genome.
    if params["MUTATION_PROBABILITY"] and params["MUTATION_EVENTS"] == 1:
        p_mut = params["MUTATION_PROBABILITY"]
    elif params["MUTATION_PROBABILITY"] and params["MUTATION_EVENTS"] > 1:
        s = (
            "operators.mutation.int_flip_per_codon\n"
            "Error: mutually exclusive parameters for 'MUTATION_PROBABILITY'"
            "and 'MUTATION_EVENTS' have been explicitly set.\n"
            "       Only one of these parameters can be used at a time with"
            "int_flip_per_codon mutation."
        )
        raise Exception(s)
    else:
        # Default mutation events per individual is 1. Raising this number
        # will influence the mutation probability for each codon.
        p_mut = params["MUTATION_EVENTS"] / eff_length

    # Mutation probability works per-codon over the portion of the
    # genome as defined by the within_used flag.
    for i in range(eff_length):
        if random() < p_mut:
            ind.genome[i] = randint(0, params["CODON_SIZE"])

    # Re-build a new individual with the newly mutated genetic information.
    new_ind = individual.Individual(ind.genome, None)

    return new_ind


def int_flip_per_ind(ind):
    """
    Mutate the genome of an individual by randomly choosing a new int with
    probability p_mut. Works per-individual. Mutation is performed over the
    entire length of the genome by default, but the flag within_used is
    provided to limit mutation to only the effective length of the genome.

    :param ind: An individual to be mutated.
    :return: A mutated individual.
    """

    # Set effective genome length over which mutation will be performed.
    eff_length = get_effective_length(ind)

    if not eff_length:
        # Linear mutation cannot be performed on this individual.
        return ind

    for _ in range(params["MUTATION_EVENTS"]):
        idx = randint(0, eff_length - 1)
        ind.genome[idx] = randint(0, params["CODON_SIZE"])

    # Re-build a new individual with the newly mutated genetic information.
    new_ind = individual.Individual(ind.genome, None)

    return new_ind


def subtree(ind):
    """
    Mutate the individual by replacing a randomly selected subtree with a
    new randomly generated subtree. Guaranteed one event per individual, unless
    params['MUTATION_EVENTS'] is specified as a higher number.

    :param ind: An individual to be mutated.
    :return: A mutated individual.
    """

    def subtree_mutate(ind_tree):
        """
        Creates a list of all nodes and picks one node at random to mutate.
        Because we have a list of all nodes, we can (but currently don't)
        choose what kind of nodes to mutate on. Handy.

        :param ind_tree: The full tree of an individual.
        :return: The full mutated tree and the associated genome.
        """

        # Find the list of nodes we can mutate from.
        targets = ind_tree.get_target_nodes(
            [], target=params["BNF_GRAMMAR"].non_terminals
        )

        # Pick a node.
        new_tree = choice(targets)

        # Set the depth limits for the new subtree.
        if params["MAX_TREE_DEPTH"]:
            # Set the limit to the tree depth.
            max_depth = params["MAX_TREE_DEPTH"] - new_tree.depth

        else:
            # There is no limit to tree depth.
            max_depth = None

        # Mutate a new subtree.
        generate_tree(new_tree, [], [], "random", 0, 0, 0, max_depth)

        return ind_tree

    if ind.invalid:
        # The individual is invalid.
        tail = []

    else:
        # Save the tail of the genome.
        tail = ind.genome[ind.used_codons:]

    # Allows for multiple mutation events should that be desired.
    for i in range(params["MUTATION_EVENTS"]):
        ind.tree = subtree_mutate(ind.tree)

    # Re-build a new individual with the newly mutated genetic information.
    ind = individual.Individual(None, ind.tree)

    # Add in the previous tail.
    ind.genome = ind.genome + tail

    return ind


def get_effective_length(ind):
    """
    Return the effective length of the genome for linear mutation.

    :param ind: An individual.
    :return: The effective length of the genome.
    """

    if not ind.genome:
        # The individual does not have a genome; linear mutation cannot be
        # performed.
        return None

    elif ind.invalid:
        # Individual is invalid.
        eff_length = len(ind.genome)

    elif params["WITHIN_USED"]:
        eff_length = min(len(ind.genome), ind.used_codons)

    else:
        eff_length = len(ind.genome)

    return eff_length


def LTGE_mutation(ind):
    """Mutation in the LTGE representation."""

    # mutate and repair.
    g, ph = latent_tree_repair(
        latent_tree_mutate(
            ind.genome), params["BNF_GRAMMAR"], params["MAX_TREE_DEPTH"]
    )

    # wrap up in an Individual and fix up various Individual attributes
    ind = individual.Individual(g, None, False)

    ind.phenotype = ph

    # number of nodes is the number of decisions in the genome
    ind.nodes = ind.used_codons = len(g)

    # each key is the length of a path from root
    ind.depth = max(len(k) for k in g)

    # in LTGE there are no invalid individuals
    ind.invalid = False

    return ind


# Set attributes for all operators to define linear or subtree representations.
int_flip_per_codon.representation = "linear"
int_flip_per_ind.representation = "linear"
subtree.representation = "subtree"
LTGE_mutation.representation = "latent tree"



from math import sqrt

from autooc.algorithm.fitness.math_functions import binary_phen_to_float
from autooc.fitness.base_ff_classes.base_ff import base_ff


class zdt1(base_ff):
    """
    Fitness function for the first problem (T_1) presented in
    [Zitzler2000].

    .. Zitzler, Eckart, Kalyanmoy Deb, and Lothar Thiele. Comparison
    of multiobjective evolutionary algorithms: Empirical results.
    Evolutionary computation 8.2 (2000): 173-195.
    """

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

    def evaluate(self, ind, **kwargs):

        min_value = [0] * 30
        max_value = [1] * 30

        real_chromosome = binary_phen_to_float(
            ind.phenotype, 30, min_value, max_value)

        summation = 0
        for i in range(1, len(real_chromosome)):
            summation += real_chromosome[i]

        g = 1 + 9 * summation / (len(real_chromosome) - 1.0)
        h = 1 - sqrt(real_chromosome[0] / g)

        return g * h



import numpy as np

from autooc.algorithm.parameters import params
from autooc.representation import individual


def check_ind(ind, check):
    """
    Check all shallow aspects of an individual to ensure everything is correct.

    :param ind: An individual to be checked.
    :return: False if everything is ok, True if there is an issue.
    """

    if ind.genome == []:
        # Ensure all individuals at least have a genome.
        return True

    if ind.invalid and (
        (check == "crossover" and params["NO_CROSSOVER_INVALIDS"])
        or (check == "mutation" and params["NO_MUTATION_INVALIDS"])
    ):
        # We have an invalid.
        return True

    elif params["MAX_TREE_DEPTH"] and ind.depth > params["MAX_TREE_DEPTH"]:
        # Tree is too deep.
        return True

    elif params["MAX_TREE_NODES"] and ind.nodes > params["MAX_TREE_NODES"]:
        # Tree has too many nodes.
        return True

    elif params["MAX_GENOME_LENGTH"] and len(ind.genome) > params["MAX_GENOME_LENGTH"]:
        # Genome is too long.
        return True


def check_genome_mapping(ind):
    """
    Re-maps individual to ensure genome is correct, i.e. that it maps to the
    correct phenotype and individual.

    :param ind: An instance of the representation.individual.Individual class.
    :return: Nothing.
    """

    # Re-map individual using fast genome mapper to check everything is ok
    new_ind = individual.Individual(ind.genome, None)

    # Get attributes of both individuals.
    attributes_0 = vars(ind)
    attributes_1 = vars(new_ind)

    if params["GENOME_OPERATIONS"]:
        # If this parameter is set then the new individual will have no tree.
        attributes_0["tree"] = None

    else:
        if attributes_0["tree"] != attributes_1["tree"]:
            s = (
                "utilities.representation.check_methods.check_ind.\n"
                "Error: Individual trees do not match."
            )
            raise Exception(s)

    # Check that all attributes match across both individuals.
    for a_0 in sorted(attributes_0.keys()):
        for a_1 in sorted(attributes_1.keys()):
            if (
                a_0 == a_1
                and attributes_0[a_0] != attributes_1[a_1]
                and not (
                    type(attributes_0[a_0]) is float
                    and type(attributes_1[a_1]) is float
                    and np.isnan(attributes_0[a_0])
                    and np.isnan(attributes_1[a_1])
                )
            ):

                s = (
                    "utilities.representation.check_methods."
                    "check_genome_mapping\n"
                    "Error: Individual attributes do not match genome-"
                    "encoded attributes.\n"
                    "       Original attribute:\n"
                    "           %s :\t %s\n"
                    "       Encoded attribute:\n"
                    "           %s :\t %s"
                    % (a_0, attributes_0[a_0], a_1, attributes_1[a_1])
                )
                raise Exception(s)


def check_ind_from_parser(ind, target):
    """
    Checks the mapping of an individual generated by the GE parser against
    the specified target string to ensure the GE individual is correct.

    :param ind: An instance of the representation.individual.Individual class.
    :param target: A target string against which to match the phenotype of
    the individual.
    :return: Nothing.
    """

    # Re-map individual using genome mapper to check everything is ok.
    new_ind = individual.Individual(ind.genome, None)

    # Check phenotypes are the same.
    if new_ind.phenotype != ind.phenotype:
        s = (
            "utilities.representation.check_methods.check_ind_from_parser\n"
            "Error: Solution phenotype doesn't match genome mapping.\n"
            "       Solution phenotype:  \t %s\n"
            "       Solution from genome:\t %s\n"
            "       Derived genome:      \t %s"
            % (ind.phenotype, new_ind.phenotype, ind.genome)
        )
        raise Exception(s)

    # Check the phenotype matches the target string.
    elif ind.phenotype != target:
        s = (
            "utilities.representation.check_methods.check_ind_from_parser\n"
            "Error: Solution phenotype doesn't match target.\n"
            "       Target:   \t %s\n"
            "       Solution: \t %s" % (target, ind.phenotype)
        )
        raise Exception(s)

    else:
        # Check the tree matches the phenotype.
        check_genome_mapping(ind)


def check_genome_from_tree(ind_tree):
    """
    Goes through a tree and checks each codon to ensure production choice is
    correct.

    :param ind_tree: The representation.tree.Tree class derivation tree of
    an individual.
    :return: Nothing.
    """

    if ind_tree.children:
        # This node has children and thus must have an associated codon.

        if not ind_tree.codon:
            s = (
                "utilities.representation.check_methods."
                "check_genome_from_tree\n"
                "Error: Node with children has no codon.\n"
                "       %s" % (str(ind_tree.children))
            )
            raise Exception(s)

        # Check production choices for node root.
        productions = params["BNF_GRAMMAR"].rules[ind_tree.root]["choices"]

        # Select choice based on node codon.
        selection = ind_tree.codon % len(productions)
        chosen_prod = productions[selection]

        # Build list of roots of the chosen production.
        prods = [prod["symbol"] for prod in chosen_prod["choice"]]
        roots = []

        # Build list of the roots of all node children.
        for kid in ind_tree.children:
            roots.append(kid.root)

        # Match production roots with children roots.
        if roots != prods:
            s = (
                "utilities.representation.check_methods."
                "check_genome_from_tree\n"
                "Error: Codons are incorrect for given tree.\n"
                "       Codon productions:\t%s\n       "
                "       Actual children:\t%s" % (str(prods), str(roots))
            )
            raise Exception(s)

    for kid in ind_tree.children:
        # Recurse over all children.
        check_genome_from_tree(kid)


def check_expansion(tree, nt_keys):
    """
    Check if a given tree is completely expanded or not. Return boolean
    True if the tree IS NOT completely expanded, i.e. if tree is invalid.

    :param tree: An individual's derivation tree.
    :param nt_keys: The list of all non-terminals.
    :return: True if tree is not fully expanded, else False.
    """

    check = False
    if tree.root in nt_keys:
        # Current node is a NT and should have children
        if tree.children:
            # Everything is as expected
            for child in tree.children:
                # Recurse over all children.
                check = child.check_expansion(nt_keys)

                if check:
                    # End recursion.
                    break

        else:
            # Current node is not completely expanded.
            check = True

    return check


def build_genome(tree, genome):
    """
    Goes through a tree and builds a genome from all codons in the subtree.

    :param tree: An individual's derivation tree.
    :param genome: The list of all codons in a subtree.
    :return: The fully built genome of a subtree.
    """

    if tree.codon:
        # If the current node has a codon, append it to the genome.
        genome.append(tree.codon)

    for child in tree.children:
        # Recurse on all children.
        genome = child.build_genome(genome)

    return genome


def get_nodes_and_depth(tree, nodes=0, max_depth=0):
    """
    Get the number of nodes and the max depth of the tree.

    :param tree: An individual's derivation tree.
    :param nodes: The number of nodes in a tree.
    :param max_depth: The maximum depth of any node in the tree.
    :return: number, max_depth.
    """

    # Increment number of nodes in the tree.
    nodes += 1

    # Set the depth of the current node.
    if tree.parent:
        tree.depth = tree.parent.depth + 1
    else:
        tree.depth = 1

    # Check the recorded max_depth.
    if tree.depth > max_depth:
        max_depth = tree.depth

    # Create list of all non-terminal children of current node.
    NT_kids = [
        kid for kid in tree.children if kid.root in params["BNF_GRAMMAR"].non_terminals
    ]

    if not NT_kids and get_output(tree):
        # Current node has only terminal children.
        nodes += 1

        # Terminal children increase the current node depth by one.
        # Check the recorded max_depth.
        if tree.depth + 1 > max_depth:
            max_depth = tree.depth + 1

    else:
        for child in NT_kids:
            # Recurse over all children.
            nodes, max_depth = get_nodes_and_depth(child, nodes, max_depth)

    return nodes, max_depth


def get_max_tree_depth(tree, max_depth=1):
    """
    Returns the maximum depth of the tree from the current node.

    :param tree: The tree we wish to find the maximum depth of.
    :param max_depth: The maximum depth of the tree.
    :return: The maximum depth of the tree.
    """

    curr_depth = get_current_depth(tree)
    if curr_depth > max_depth:
        max_depth = curr_depth
    for child in tree.children:
        max_depth = get_max_tree_depth(child, max_depth)
    return max_depth


def get_current_depth(tree):
    """
    Get the depth of the current node by climbing back up the tree until no
    parents remain (i.e. the root node has been reached).

    :param tree: An individual's derivation tree.
    :return: The depth of the current node.
    """

    # Set the initial depth at 1.
    depth = 1

    # Set the current parent.
    current_parent = tree.parent

    while current_parent is not None:
        # Recurse until the root node of the tree has been reached.

        # Increment depth.
        depth += 1

        # Set new parent.
        current_parent = current_parent.parent

    return depth


def get_output(ind_tree):
    """
    Calls the recursive build_output(self) which returns a list of all
    node roots. Joins this list to create the full phenotype of an
    individual. This two-step process speeds things up as it only joins
    the phenotype together once rather than at every node.

    :param ind_tree: a full tree for which the phenotype string is to be built.
    :return: The complete built phenotype string of an individual.
    """

    def build_output(tree):
        """
        Recursively adds all node roots to a list which can be joined to
        create the phenotype.

        :return: The list of all node roots.
        """

        output = []
        for child in tree.children:
            if not child.children:
                # If the current child has no children it is a terminal.
                # Append it to the output.
                output.append(child.root)

            else:
                # Otherwise it is a non-terminal. Recurse on all
                # non-terminals.
                output += build_output(child)

        return output

    return "".join(build_output(ind_tree))


def ret_true(obj):
    """
    Returns "True" if an object is there. E.g. if given a list, will return
    True if the list contains some data, but False if the list is empty.

    :param obj: Some object (e.g. list)
    :return: True if something is there, else False.
    """

    if obj:
        return True
    else:
        return False


def generate_codon(NT, choice):
    """
    Given a list of choices and a choice from that list, generate and return a
    codon which will result in that production choice being made.

    :param NT: A root non-terminal node from which production choices are made.
    :param choice: A production choice from the available choices of the
    given NT.
    :return: A codon that will give that production choice.
    """

    productions = params["BNF_GRAMMAR"].rules[NT]

    # Find the production choices from the given NT.
    choices = [choice["choice"] for choice in productions["choices"]]

    # Find the index of the chosen production and set a matching codon based
    # on that index.
    prod_index = choices.index(choice)

    codon = productions["no_choices"] + prod_index

    # Generate a valid codon.
    return codon


def check_tree(tree):
    """
    Recursively traverse a tree and ensure that all parents and children are
    correct.

    :param tree: A tree.
    :return: Nothing.
    """

    if tree.children:

        if not tree.codon:
            s = (
                "utilities.representation.check_methods.check_tree\n"
                "Error: Node with children has no associated codon."
            )
            raise Exception(s)

        for child in tree.children:

            if child.parent != tree:
                s = (
                    "utilities.representation.check_methods.check_tree\n"
                    "Error: Child doesn't belong to parent.\n"
                    "       Child parent:  %s\n"
                    "       Actual parent: %s\n"
                    "       Child P depth: %s\n"
                    "       Parent depth:  %s"
                    % (child.parent.root, tree.root, child.parent.depth, tree.depth)
                )
                raise Exception(s)

            else:
                check_tree(child)



import json
import time
from os import path
from xml.dom.pulldom import parseString

import numpy as np
from sklearn.ensemble import IsolationForest
from sklearn.neighbors import LocalOutlierFactor
from sklearn.svm import OneClassSVM
from tensorflow.keras import Input, Sequential
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.layers import BatchNormalization, Dense, Dropout, Input, Lambda
from tensorflow.keras.utils import plot_model

from autooc.algorithm.parameters import params
from autooc.fitness.base_ff_classes.base_ff import base_ff
from autooc.stats.stats import stats
from autooc.utilities.fitness.error_metric import (auc_autoencoder, auc_sklearn,
                                                  reconstruction_error)
from autooc.utilities.fitness.get_data import get_data
from autooc.utilities.fitness.optimize_constants import optimize_constants
from autooc.utilities.utils import (get_model_from_encoder,
                                   get_model_from_encoder2)

np.seterr(all="raise")


class supervised_learning(base_ff):
    """
    Fitness function for supervised learning, ie regression and
    classification problems. Given a set of training or test data,
    returns the error between y (true labels) and yhat (estimated
    labels).

    We can pass in the error metric and the dataset via the params
    dictionary. Of error metrics, eg RMSE is suitable for regression,
    while F1-score, hinge-loss and others are suitable for
    classification.

    This is an abstract class which exists just to be subclassed:
    should not be instantiated.
    """

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

        # Get training and test data
        # self.training_in, self.training_exp, self.test_in, self.test_exp = \
        # self.dtrain, self.dval = params['normal_train_data'], params['test_data']
        self.dtrain, self.dval = params["X_train"], params["X_val"]
        # get_data(params['DATASET_TRAIN'], params['DATASET_TEST'])

    def evaluate(self, ind, **kwargs):
        """
        Note that math functions used in the solutions are imported from either
        utilities.fitness.math_functions or called from numpy.

        :param ind: An individual to be evaluated.
        :param kwargs: An optional parameter for problems with training/test
        data. Specifies the distribution (i.e. training or test) upon which
        evaluation is to be performed.
        :return: The fitness of the evaluated individual.
        """

        if params["ALGORITHM"] in ["autoencoder", "vae", "nas"]:
            metric = evaluate_autoencoder(ind, self.dtrain)
        elif params["ALGORITHM"] in ["iforest", "svm", "lof"]:
            metric = evaluate_sklearn(ind, self.dtrain)
        elif params["ALGORITHM"] in "all":
            metric = evaluate_all(ind, self.dtrain)

        return metric


def evaluate_autoencoder(ind, X_train):
    shape = X_train.shape[1]

    d = {
        "Sequential": Sequential,
        "Dense": Dense,
        "Lambda": Lambda,
        "BatchNormalization": BatchNormalization,
        "Input": Input,
        "Dropout": Dropout,
        "get_model_from_encoder": get_model_from_encoder2,
        "input_shape": shape,
    }

    exec(ind.phenotype, d)
    print("FENOTIPO")
    print(ind.phenotype)
    print("END FENOTIPO")
    #print(type(ind.phenotype))
    #print("\nSTART D:\n")
    #print(d)
    #print("\nEND D:\n")

    model = d["model"]

    # print("MODELO:")
    # model = get_model_from_encoder(ind.phenotype)
    # print(model)
    early_stop = EarlyStopping(
        monitor="val_loss",
        patience=3,
        verbose=1,
        mode="min",
        restore_best_weights=True,
    )
    fit_start = time.time()
    hist = model.fit(
        X_train,
        X_train,
        epochs=params["EPOCHS"],
        # validation_data=(self.dval, self.dval),
        validation_split=0.25,
        verbose=1,
        callbacks=[early_stop],
    )
    fit_end = time.time()

    training_time = fit_end - fit_start
    ind.training_time = training_time

    predict_start = time.time()
    single_pred = model.predict(X_train[0:1])
    predict_end = time.time()
    predict_time = predict_end - predict_start

    ind.predict_time = predict_time
    ind.num_params = model.count_params()
    ind.model = model

    # save model history
    rand_name = str(stats["gen"]) + "_" + str(ind.name) + "_hist.json"
    filename = path.join(params["FILE_PATH"], rand_name)
    with open(filename, "w") as f:
        json.dump(hist.history, f)

    # save model plot
    print(model.summary())
    rand_name = str(stats["gen"]) + "_" + str(ind.name) + "_model.png"
    filename = path.join(params["FILE_PATH"], rand_name)
    plot_model(
        model, filename, show_shapes=True, show_dtype=True, expand_nested=True
    )

    if params["TYPE"] == "unsupervised":
        metric, rec_losses, threshold = metric_rec_error(model, ind)
    else:
        metric, rec_error, rec_losses, threshold = metric_auc(model, ind)

    ind.threshold = threshold
    ind.losses = rec_losses
    ind.rec_error = metric

    # save complexity scores
    filename = path.join(params["FILE_PATH"], "complexity.csv")
    with open(filename, "a") as the_file:
        the_file.write(
            (
                str(stats["gen"])
                + ";"
                + str(ind.name)
                + ";"
                + str(training_time)
                + "\n"
            )
        )

    #print(metric)
    return metric


def evaluate_sklearn(ind, X_train, algorithm=None):
    if params["ALGORITHM"] == "iforest" or algorithm == "iforest":
        d = {"IsolationForest": IsolationForest}
    elif params["ALGORITHM"] == "svm" or algorithm == "svm":
        d = {"OneClassSVM": OneClassSVM}
    elif params["ALGORITHM"] == "lof" or algorithm == "lof":
        d = {"LocalOutlierFactor": LocalOutlierFactor}

    exec(ind.phenotype, d)
    print("FENOTIPO")
    print(ind.phenotype)
    print("END FENOTIPO")
    #print(type(ind.phenotype))
    #print("\nSTART D:\n")
    #print(d)
    #print("\nEND D:\n")

    model = d["model"]

    fit_start = time.time()
    model.fit(X_train)
    fit_end = time.time()

    if params["TYPE"] == "unsupervised":
        metric = metric_anomaly_score(model, ind)
    else:
        metric = metric_auc_sklearn(model, ind)

    training_time = fit_end - fit_start
    ind.training_time = training_time

    predict_start = time.time()
    single_pred = model.predict(X_train[0:1])
    predict_end = time.time()
    predict_time = predict_end - predict_start

    ind.predict_time = predict_time
    #ind.num_params = model.count_params()
    ind.model = model
    ind.generation = stats["gen"]

    # save complexity scores
    filename = path.join(params["FILE_PATH"], "complexity.csv")
    with open(filename, "a") as the_file:
        the_file.write(
            (
                str(stats["gen"])
                + ";"
                + str(ind.name)
                + ";"
                + str(ind.id)
                + ";"
                + str(training_time)
                + "\n"
            )
        )

    #print(metric)
    return metric


def evaluate_all(ind, X_train):
    shape = X_train.shape[1]
    d = {
        "Sequential": Sequential,
        "Dense": Dense,
        "Lambda": Lambda,
        "BatchNormalization": BatchNormalization,
        "Input": Input,
        "Dropout": Dropout,
        "get_model_from_encoder": get_model_from_encoder2,
        "input_shape": shape,
        "OneClassSVM": OneClassSVM,
        "IsolationForest": IsolationForest,
        "LocalOutlierFactor": LocalOutlierFactor
    }

    exec(ind.phenotype, d)
    print("FENOTIPO")
    print(ind.phenotype)
    print("END FENOTIPO")
    # print(type(ind.phenotype))
    #print("\nSTART D:\n")
    #print(d)
    #print("\nEND D:\n")

    model = d["model"]
    #print(dir(model))
    # print(type(model.__class__.__name__))

    if model.__class__.__name__ == "Sequential":
        metric = evaluate_autoencoder(ind, X_train)
    elif model.__class__.__name__ == "OneClassSVM":
        metric = evaluate_sklearn(ind, X_train, algorithm="svm")
    elif model.__class__.__name__ == "IsolationForest":
        metric = evaluate_sklearn(ind, X_train, algorithm="iforest")
    elif model.__class__.__name__ == "LocalOutlierFactor":
        metric = evaluate_sklearn(ind, X_train, algorithm="lof")

    return metric


def metric_auc(model, ind):
    # calculate reconstruction error
    #auc = params["ERROR_METRIC"](model)
    auc, rec_error, rec_losses, threshold = auc_autoencoder(model)

    # save prediction scores
    filename = path.join(params["FILE_PATH"], "val_aucs.csv")
    with open(filename, "a") as the_file:
        the_file.write(
            (str(stats["gen"]) + ";" + str(ind.name) +
                ";" + str(ind.id) + ";" + str(auc) + "\n")
        )

    return auc, rec_error, rec_losses, threshold


def metric_rec_error(model, ind):
    # calculate reconstruction error
    rec_error, rec_losses, threshold = params["ERROR_METRIC"](model)

    # save prediction scores
    filename = path.join(params["FILE_PATH"], "val_rec_errors.csv")
    with open(filename, "a") as the_file:
        the_file.write(
            (str(stats["gen"]) + ";" + str(ind.name) +
                ";" + str(ind.id) +
                ";" + str(rec_error) + "\n")
        )

    return rec_error, rec_losses, threshold


def metric_auc_sklearn(model, ind):
    # calculate reconstruction error
    # auc = params["ERROR_METRIC"](model)
    auc = auc_sklearn(model)

    # save prediction scores
    filename = path.join(params["FILE_PATH"], "val_aucs.csv")
    with open(filename, "a") as the_file:
        the_file.write(
            (str(stats["gen"]) + ";" + str(ind.name) +
                ";" + str(ind.id) +
                ";" + str(auc) + "\n")
        )

    return auc


def metric_anomaly_score(model, ind):
    # calculate anomaly score
    #auc = params["ERROR_METRIC"](model)
    auc = params["ERROR_METRIC"](model)

    # save prediction scores
    filename = path.join(params["FILE_PATH"], "val_anomaly_scores.csv")
    with open(filename, "a") as the_file:
        the_file.write(
            (str(stats["gen"]) + ";" + str(ind.name) +
                ";" + str(ind.id) + ";" + str(auc) + "\n")
        )

    return auc



import getopt
import sys
from os import getcwd, listdir, path, sep
from sys import path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from autooc.algorithm.algorithm.general import check_python_version

path.append("../src")


check_python_version()


np.seterr(all="raise")

matplotlib.use("Agg")

plt.rc("font", family="Times New Roman")


def help_message():
    """
    Prints a help message to explain the usage of this file.

    :return: Nothing
    """

    lines_1 = [
        "Welcome to PonyGE's post-run stats parser.",
        "-------------------",
        "The following are the available command line args.",
        "You must specify an experiment name.",
    ]

    lines_2 = [
        ["\t--help:", "Shows this help message."],
        [
            "\t--experiment_name:",
            "The name of the containing folder in "
            "which target runs are saved, e.g. "
            "in results/[EXPERIMENT_NAME].",
        ],
    ]

    # This simply justifies the print statement such that it is visually
    # pleasing to look at.
    for line in lines_1:
        print(line)
    col_width = max(len(line[0]) for line in lines_2)
    for line in sorted(lines_2):
        print(" ".join(words.ljust(col_width) for words in line))


def parse_opts(command_line_args):
    """
    Parses command line arguments and returns usable variables which are used
    as inputs for other functions in this file.

    :param command_line_args: flags passed in from the command line at
    execution
    :return: experiment_name: the name of the containing folder of results,
             stats: a list of strings. Each string is a valid key from the
                    stats.stats.stats dictionary.
             graph: an optional boolean flag for graphing specified stats.
    """

    try:
        opts, args = getopt.getopt(
            command_line_args[1:], "", ["help", "experiment_name="]
        )
    except getopt.GetoptError as err:
        s = (
            "scripts.parse_stats.parse_opts\n"
            "Error: in order to parse stats you need to specify the location"
            " of the target stats files.\n"
            "       Run python stats_parser.py --help for more info."
        )
        print(str(err))
        raise Exception(s)

    if not opts:
        s = (
            "scripts.parse_stats.parse_opts\n"
            "Error: in order to parse stats you need to specify the location"
            " of the target stats files.\n"
            "       Run python stats_parser.py --help for more info."
        )
        raise Exception(s)

    experiment_name = None

    # iterate over all arguments in the option parser.
    for opt, arg in opts:
        if opt == "--help":
            # Print help message.
            help_message()
            exit()

        elif opt == "--experiment_name":
            # Set experiment name (i.e. containing folder for multiple runs).
            experiment_name = arg

    return experiment_name


def parse_stats_from_runs(experiment_name):
    """
    Analyses a list of given stats from a group of runs saved under an
    "experiment_name" folder. Creates a summary .csv file which can be used by
    plotting functions in utilities.save_plot. Saves a file of the format:

        run0_gen0       run1_gen0       .   .   .   run(n-1)_gen0
        run0_gen1       run1_gen1       .   .   .   run(n-1)_gen1
        run0_gen2       run1_gen2       .   .   .   run(n-1)_gen2
        .               .               .   .   .   .
        .               .               .   .   .   .
        .               .               .   .   .   .
        run0_gen(n-1)   run1_gen(n-1)   .   .   .   run(n-1)_gen(n-1)
        run0_gen(n)     run1_gen(n)     .   .   .   run(n-1)_gen(n)

    Generated file is compatible with

        utilities.save_plot.save_average_plot_across_runs()

    :param experiment_name: The name of a collecting folder within the
    ./results folder which holds multiple runs.
    :param graph: A boolean flag for whether or not to save figure.
    :return: Nothing.
    """

    # Since results files are not kept in source directory, need to escape
    # one folder.
    file_path = path.join(getcwd(), "autooc", "results")

    # Check for use of experiment manager.
    if experiment_name:
        file_path = path.join(file_path, experiment_name)

    else:
        s = (
            "scripts.parse_stats.parse_stats_from_runs\n"
            "Error: experiment name not specified."
        )
        raise Exception(s)

    # Find list of all runs contained in the specified folder.
    runs = [run for run in listdir(
        file_path) if path.isdir(path.join(file_path, run))]

    # Place to store the header for full stats file.
    header = ""

    # Array to store all stats
    full_stats = []

    # Get list of all stats to parse. Check stats file of first run from
    # runs folder.
    ping_file = path.join(file_path, str(runs[0]), "stats.tsv")

    # Load in data and get the names of all stats.
    stats = list(pd.read_csv(ping_file, sep="\t"))

    # Make list of stats we do not wish to parse.
    no_parse_list = ["gen", "total_inds", "time_adjust"]

    for stat in [
        stat
        for stat in stats
        if stat not in no_parse_list and not stat.startswith("Unnamed")
    ]:
        # Iterate over all stats.
        print("Parsing", stat)
        summary_stats = []

        # Iterate over all runs
        for run in runs:
            # Get file name
            file_name = path.join(file_path, str(run), "stats.tsv")

            # Load in data
            data = pd.read_csv(file_name, sep="\t")

            try:
                # Try to extract specific stat from the data.
                if list(data[stat]):
                    summary_stats.append(list(data[stat]))
                else:
                    s = (
                        "scripts.parse_stats.parse_stats_from_runs\n"
                        "Error: stat %s is empty for run %s." % (stat, run)
                    )
                    raise Exception(s)

            except KeyError:
                # The requested stat doesn't exist.
                s = (
                    "scripts.parse_stats.parse_stats_from_runs\nError: "
                    "stat %s does not exist in run %s." % (stat, run)
                )
                raise Exception(s)

        try:
            # Generate numpy array of all stats
            summary_stats = np.array(summary_stats)

            # Append Stat to header.
            header = header + stat + "_mean,"

            summary_stats_mean = np.nanmean(summary_stats, axis=0)
            full_stats.append(summary_stats_mean)

            # Append Stat to header.
            header = header + stat + "_std,"
            summary_stats_std = np.nanstd(summary_stats, axis=0)
            full_stats.append(summary_stats_std)
            summary_stats = np.transpose(summary_stats)

            # Save stats as a .csv file.
            np.savetxt(
                path.join(file_path, (stat + ".csv")), summary_stats, delimiter=","
            )

            # Graph stat by calling graphing function.
            save_average_plot_across_runs(
                path.join(file_path, (stat + ".csv")))

        except FloatingPointError:
            print(
                "scripts.stats_parser.parse_stats_from_runs\n"
                "Warning: FloatingPointError encountered while parsing %s "
                "stats." % (stat)
            )

    # Convert and rotate full stats
    full_stats = np.array(full_stats)
    full_stats = np.transpose(full_stats)

    # Save full stats to csv file.
    np.savetxt(
        path.join(file_path, "full_stats.csv"),
        full_stats,
        delimiter=",",
        header=header[:-1],
    )


def save_average_plot_across_runs(filename):
    """
    Saves an average plot of multiple runs. Input file data must be of the
    format:

        run0_gen0       run1_gen0       .   .   .   run(n-1)_gen0
        run0_gen1       run1_gen1       .   .   .   run(n-1)_gen1
        run0_gen2       run1_gen2       .   .   .   run(n-1)_gen2
        .               .               .   .   .   .
        .               .               .   .   .   .
        .               .               .   .   .   .
        run0_gen(n-1)   run1_gen(n-1)   .   .   .   run(n-1)_gen(n-1)
        run0_gen(n)     run1_gen(n)     .   .   .   run(n-1)_gen(n)

    The required file can be generated using

        stats.parse_stats.parse_stats_from_runs()

    Generates a .pdf graph of average value with standard deviation.
    :param filename: the full file name of a .csv file containing the fitnesses
    of each generation of multiple runs. Must be comma separated.
    :return: Nothing.
    """

    # Get stat name from the filename. Used later for saving graph.
    stat_name = filename.split(sep)[-1].split(".")[0]

    # Load in data.
    data = np.genfromtxt(filename, delimiter=",")[:, :-1]

    # Generate average and standard deviations of loaded data.
    ave = np.nanmean(data, axis=1)
    std = np.nanstd(data, axis=1)

    # Calculate max and min of standard deviation.
    stdmax = ave + std
    stdmin = ave - std

    # Generate generation range over which data is to be graphed.
    max_gens = len(ave)
    r = range(1, max_gens + 1)

    # Initialise figure plot.
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)

    # Plot data and standard deviation infill.
    ax1.plot(r, ave, color="blue")
    ax1.fill_between(r, stdmin, stdmax, color="DodgerBlue", alpha=0.5)

    # Set x-axis limits.
    plt.xlim(0, max_gens + 1)

    # Set title and axes.
    plt.title("Average " + stat_name)
    plt.xlabel("Generation", fontsize=14)
    plt.ylabel("Average " + stat_name, fontsize=14)

    # Save graph under the same name as the original .csv file but with a
    # .pdf extension instead.
    new_filename = filename[:-3] + "pdf"
    plt.savefig(str(new_filename))

    plt.close()


if __name__ == "__main__":
    # Get experiment name and graphing flag from command line parser.
    experiment_name = parse_opts(sys.argv)

    # Call statistics parser for experiment name.
    parse_stats_from_runs(experiment_name)
